{ TAudioVolume

  This component controls volume level or muting state of audio device and Session.

  Any suggestions/advice would be appreciated.

  Author : Silhwan Hyun  (hyunsh@hanafos.com)
  Contributor : Emil Weiss (weiss.em@arcor.de)
  Contributor : HolgerX (https://www.delphipraxis.net/members/19009-holgerx.html)

  #### Copyright Notice ####
  Copyright 2011 Silhwan Hyun, All Rights Reserved
  This unit is free. It may be used both in commercial and non-commercial software either in
  original or in modified form.
  This unit can be freely distributed in any way and by any means provided this copyright
  notice is preserved.


  note) 1. For the functions, procedures and events which constitute pairs, e.g. GetVolume and GetMasterVolume,
  - The one which has "Master" in its name is related to audio endpoint devices such as speaker or
  headphone and the other is related to the application to which the instance of TAudioVolume belongs.
  2. The property SuppressEvent can be used to inhibit redundant triggering of OnVolumeEvent or OnMasterVolumeEvent.
  => Set when the thumb of TrackBar(used as Volume Level Indicator on main form) is pressed
  and reset when the thumb of TrackBar(used as Volume Level Indicator) is released.
  3. Following proceure/function/property is valid only if the operating system is Windows Vista or newer.
  IsMuted, GetVolume, DeviceMixFormat
  4. function SetupPeakMeter is valid only if the operating system is Windows 7 or newer.
  5. You should call the function SetupPeakMeter once before you use function GetPeak to get the peak value
  of application's audio output, i.e., not of the peak value of endpoint audio device.

  If you have any idea on this issue, please let me know.

  Revision History
  ------------------------------
  v1.0.2 (08 Mai 2018)
  - Caption der GroupBox entfernt.
  - Neuen Label addiert fr Anwendungs Info
  - Alle Exe Dateien werden jetzt durch eine VersionInfo Routine gejagt
    wenn vorhanden wird Filedescription anstelle des Exenamen verwendet.
    Der Quelltext dafr ist nicht von mir habe aber den Credit in den Pass Dateien erhalten es kann also jeder sehen woher er stammt. (DelphiDabbler)
    War mir zu viel Arbeit das jetzt selbst zu implementieren.
  - Label zentriert.
  - Namen die zu lang sind werden gekrzt wenn sich kein Leerzeichen im String befindet
  - Namen werden umgebrochen wenn Leerzeichen vorhanden sind und dieser zu lang ist (Wordwrap)
  - Die ganzen Controls dementsprechend neu ausgerichtet so das der Label passt.
  - Fehler in der procedure LoadIconPathfromResource korrigiert. (Dank geht an KodeZwerg, Dalai)
  - ToolTips addiert.
  - PeakMeter problem behoben
  - Neue Events addiert die aber leider nicht abfeuern keine ahnung warum
  - Jede menge unntige (doppelte Declarationen entfernt.
  - SystemSounds werden nun im Hauptformular angezeigt weil dieser immer vorhanden ist.
  (by Emil Weiss)

  v1.0.1 (06 Mai 2018)
  - Remove AppVolume is now a part of Session
  - Add Mixer Session
  (by Emil Weiss, HolgerX)

  v1.0.0 (25 Feb 2017)
  - Added GetMasterBalance
  - fixed some parts on GetMasterVolume depend on MixerApi
  - fixed some parts on SetMasterVolume depend on MixerApi
  - fixed AdjustMasterBalancePos
  (by Silhwan Hyun,  Emil Weiss)

  v0.9.8 (04 Feb 2017)
  - Added Balance
  (by Emil Weiss, Silhwan Hyun)

  v0.9.7           (25 Jan 2016)
  - Added Visuals to: GroupBox, Slider, Trackbars
  (by Emil Weiss)

  v0.9.6           (15 Jun 2014)
  - Added following 2 functions to get the peak value of all channels at once.
  (by suggestion of Emil Weiss)
  GetChannelsMasterPeak, GetChannelsPeak

  v0.9.5           (06 Jun 2014)
  - Added and modified functions to support application's(session) peak meter

  v0.9.4           (12 Mar 2012)
  - Modifications to support multi sound cards in a system on Windows XP
  - Replaced function AudioDeviceName with function AudioDeviceInfo
  - Added properties,
  DefaultAudioDevice : shows the default audio endpoint device for output
  OnDeviceStateChange : fires at the state change or add/removal of a device
  OnDefaultDeviceChange : fires at the change of default audio endpoint device for output
  - Changed property
  SelectedDevice : data type changed from string to TDeviceInfo(Name & Id)
  - Renamed property
  OnSessionStateEvent (renamed from OnStateEvent, state = one of Inactive, Active, Expired)

  v0.9.3           (21 Aug 2011)
  - Added functions : GetMasterPeak
  - Added property : DeviceMixFormat, MasterChannels
  - Cleaned up garbage sentences added for debug

  v0.9.2           (12 Aug 2011)
  - Bug fix : Errors caused by unnecessary call of "_Release" method of COM objects.
  - The PropSys.pas file is renamed to PropSys2.pas to avoid conflict with Delphi 2010.
  - Fixed compile error with Delphi 7.
  - The property "Active" is replaced with property "Ready" (=> it's ready to operate)
  - No selection of audio endpoint device at component Creation.

  v0.9.1           (03 Aug 2011)
  - Supports Windows XP
  - A parameter added at TOnVolumeEvent to enable you to check if the event was triggered by
  self action.

  v0.9.0           (31 Jul 2011)
  - Initial release


  Known bugs
  - There are uncertain cases which stop further event call (OnMasterVolumeEvent, OnVolumeEvent).
}

unit AudioVolume;

interface

uses
  Windows, Messages, Dialogs, ShellApi, ShlObj, SysUtils, StrUtils, Classes, ActiveX, contnrs,
  ExtCtrls, ComCtrls, Graphics, StdCtrls, propsys2, Controls, Registry, MixerAPI, MMSystem, Math,
  // MMDevice
  MMDevice,
  MMDeviceEnumerator,
  MMDeviceCollection,
  MMNotificationClient,
  // EndPointVolume
  AudioEndpointVolumeCallback,
  AudioEndpointEvents,
  AudioEndpointVolume,
  // Wasapi
  AudioClient,
  AudioSessionEvents,
  AudioSessionManager,
  AudioSessionManager2,
  AudioSessionEnumerator,
  AudioSessionControl,
  SimpleAudioVolume,
  AudioSessionControl2,
  AudioMeterInformation,
  AudioSessionNotification,
  ChannelAudioVolume,
  // Constants
  Constants,
  // VerInfo
  UVerInfoClass;

const
  // Constants fo Master Balance & SysSound Balance
  MasterBalance_Max = 32;
  MasterBalance_Mid = MasterBalance_Max div 2;
  MasterBalance_Min = 0;
  SessionBalance_Max = 32;
  SessionBalance_Mid = SessionBalance_Max div 2;
  SessionBalance_Min = 0;

type
  TAudioSessionInfo = record
    ActiveSession: string;
    DisplayName: string;
    StateSession: string;
    IconPath: string;
    SessionIdentifier: string;
    SessionDisconnected: string;
    ProcessID: string;
    PID: DWORD;
    IsSystemSoundsSession: HResult;
    SessionGuid: TGUID;
    Device: IMMDevice;
    Icon: TIcon;
  end;

type
  TDeviceInfo = record
    DeviceName: string;
    DeviceId: string;
  end;

  TAudioVolume = class;
  TSplitStrArray = array of string;

  // ByMyAction is used to check if this event is triggered by self action, i.e., my application's
  // execution of SetMasterMute, SetMasterVolume, SetSysSoundMute or SetSysSoundVolume.
  TOnMasterVolumeEvent = procedure(ByMyAction: boolean; Volume: single; Mute: boolean) of object;
  TOnSessionVolumeEvent = procedure(ByMyAction: boolean; Volume: single; Mute: boolean) of object;
  TOnSessionStateEvent = procedure(Sender: TAudioVolume; NewState: integer) of object;
  TOnSessionDisconnected = procedure(Sender: TAudioVolume; DisconnectReason: uint) of object;
  TOnDeviceStateChange = procedure(DeviceId: string; NewDeviceState: DWORD) of object;
  TOnDefaultDeviceChange = procedure(NewDefaultDevice: TDeviceInfo) of object;
  TOnSessionCreated = procedure(NewSession: IAudioSessionControl) of object;

  TAudioVolume = class(TWinControl)
  protected
    DeviceEnumerator: IMMDeviceEnumerator;
    DeviceCollection: IMMDeviceCollection;
    AudioEndpointVolume: IAudioEndpointVolume;
    AudioEndpointEvents: IAudioEndpointVolumeCallback;
    MMNotificationClient: TMMNotificationClient;
    SessionManager: IAudioSessionManager;
    SessionManager2: IAudioSessionManager2;

    // AudioClient, ChannelAudioVolume interface
    AudioClient: IAudioClient;
    ChannelAudioVolume: IChannelAudioVolume;
    SessionEnumerator: IAudioSessionEnumerator;
    SessionControl: IAudioSessionControl;
    SessionEvents: IAudioSessionEvents;
    SessionNotification: IAudioSessionNotification;
    SimpleAudioVolume: ISimpleAudioVolume;
    PeakMeter: IAudioMeterInformation;
    MasterMeter: IAudioMeterInformation;
    pbMasterMeter: TProgressBar;
    pbMasterMeterL: TProgressBar;
    pbMasterMeterR: TProgressBar;
    ckMasterMute: TCheckbox;
    tbMasterVolume: TTrackBar;
    tbMasterBalance: TTrackBar;
    pbSessionMeter: TProgressBar;
    pbSessionMeterL: TProgressBar;
    pbSessionMeterR: TProgressBar;
    tbSessionVolume: TTrackBar;
    tbSessionBalance: TTrackBar;
    ckSessionMute: TCheckbox;
    Image: TImage;

    lblMasterPeak: TLabel;
    lblMasterPeakL: TLabel;
    lblMasterPeakR: TLabel;
    lblSessionPeak: TLabel;
    lblSessionPeakL: TLabel;
    lblSessionPeakR: TLabel;
    lblMasterBalance: TLabel;
    lblSessionBalance: TLabel;
    lblMasterCaption: TLabel;
    lblSessionCaption: TLabel;

    // For Windows XP
    MasterMute: TMixerControl;
    MasterVol: TMixerControl;
    DestLineInfo: array of TDestLineInfo;
    WINMM_DEVICECHANGE: uint;

    MyVolSet: boolean;
    MyVolVal: single;
    MyMuteSet: boolean;
    MyMuteVal: boolean;
    MySessionVolSet: boolean;

    FIcon: TIcon;
    FReady: boolean;
    FMixerAPIReady: boolean;
    FMixerObj: hMixerObj;
    HMixer: hMixerObj;
    FDeviceCount: Cardinal;
    FSessionCount: integer;
    FMixFormat: TWAVEFORMATEX;
    FChannels: DWORD;
    FSelectedDevice: TDeviceInfo;
    FDefaultDevice: TDeviceInfo;
    FDeviceList: array of TDeviceInfo;
    FAudioSessionList: TAudioSessionInfo;
    FEventHandle: HWND;
    FSuppressEvent: boolean;
    FOnMasterVolumeEvent: TOnMasterVolumeEvent;
    FOnSessionVolumeEvent: TOnSessionVolumeEvent;
    FOnSessionStateEvent: TOnSessionStateEvent;
    FOnSessionCreated: TOnSessionCreated;
    FOnSessionDisconnected: TOnSessionDisconnected;
    FOnDeviceStateChange: TOnDeviceStateChange;
    FOnDefaultDeviceChange: TOnDefaultDeviceChange;
    FMasterMute: boolean;
    FSessionVolume: boolean;
    FSessionInactive: boolean;
    FPlaying: boolean;
    FAudioDevice: TStrings;
    FMGroupBox: TGroupBox;
    FSSGroupBox: TGroupBox;
    FSGroupBox: TGroupBox;
    FMasterBalance: boolean;
    FSessionBalance: boolean;
    FDevice: IMMDevice;
    GroupGuid: TGUID;
    Peaks: array of single;
    Timer1: TTimer;
  private
    procedure tbMasterVolumeChange(Sender: TObject);
    procedure tbMasterBalanceChange(Sender: TObject);
    procedure ckMasterMuteClick(Sender: TObject);
    procedure tbSessionBalanceChange(Sender: TObject);
    procedure tbSessionVolumeChange(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure ProcessMsg(var Msg: TMessage);
    procedure SetSuppressEvent(Value: boolean);
    procedure SetSessionComponents(Value: boolean);
    procedure AdjustMasterBalancePos;
    procedure AdjustSessionBalancePos;
    function SetSysItemIcon(wsItems: PWideChar): TIcon;
    procedure ckSessionMuteClick(Sender: TObject);
    function SetupDefaultAudioEndpoint(Device_Enumerator: IMMDeviceEnumerator): boolean;
    function RefreshDeviceList(Device_Enumerator: IMMDeviceEnumerator): boolean;
    procedure FillOutputLine;
    procedure SetMasterComponents;
    procedure WMSize(var Message: TWMSize);
    message WM_SIZE;
    procedure SetPlaying(const Value: boolean);
    function GetAudioDevice: TStrings;
    procedure SetAudioDevice(const Value: TStrings);
    function IIF(AValue: boolean; const ATrue: integer; const AFalse: integer): integer;
    procedure SetMasterBalance(const Value: boolean);
    procedure SetSessionBalance(const Value: boolean);
    function GetDevice: IMMDevice;
    procedure SetDevice(const Value: IMMDevice);
    function GetProcessIdFromModuleName(ModuleName: PWideChar): LongWord;
    function GetProcessPathByPID(PID: DWORD): WideString;
    function GetProcessPathByHandle(hProcess: THandle): WideString;
    function GetMasterBalance: DWORD;
    function GetMasterVolume: single;
    function GetIcon: TIcon;
    procedure SetIcon(const Value: TIcon);
    procedure SetSessionInactive(const Value: boolean);
    procedure SetDiplayNameAndIcon(PID: DWORD; Index: integer;
      out SessionAudioVolume: ISimpleAudioVolume; SessionControl: IAudioSessionControl;
      SessionControl2: IAudioSessionControl2; State: uint);
    function LoadDisplayNamefromDLL(DisplayName: PWideChar): WideString;
    procedure LoadIconPathfromResource(IconPath: PWideChar);
    function Split(const Source, Delimiter: string): TSplitStrArray;
    function GetVerInfo(FileName: WideString): WideString;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function AudioSessionInfo(SessionIndex: integer): TAudioSessionInfo;
    function AudioDeviceInfo(DeviceIndex: Cardinal): TDeviceInfo;
    function SelectAudioDevice(DeviceIndex: Cardinal): boolean;
    property SelectedDevice: TDeviceInfo read FSelectedDevice;
    function IsMasterMuted: boolean;

    function IsMuted: boolean;
    function GetVolume: single;
    function GetMasterPeak: single;
    function SetupPeakMeter(CurProcId: DWord): boolean;
    // ** Note : function "SetupPeakMeter" should be called after initialization of application.
    function GetPeak: single; // Gets application's audio output level
    function GetChannelsMasterPeak(var Channels: uint; var Peaks: array of single): boolean;
    function GetChannelsPeak(var Channels: uint; var Peaks: array of single): boolean;
    function SetSimpleAudioVolumeFromModuleName(const AudioDevice: IMMDevice;
      ModuleName: PWideChar; Index: integer; out SimpleAudio: ISimpleAudioVolume;
      out currguid: TGUID): boolean; stdcall;
    procedure GetProzessList(ProcList: TStringList);
    function SetSimpleAudioVolumeFromHandle(const AudioDevice: IMMDevice; HWND: uint;
      Index: integer; out SimpleAudio: ISimpleAudioVolume; out currguid: TGUID): boolean; stdcall;
    function SetSimpleAudioVolumeFromPid(const AudioDevice: IMMDevice; ProcId: uint;
      Index: integer; out SessionAudioVolume: ISimpleAudioVolume; out currguid: TGUID): boolean;
      stdcall;
    function GetSessionCount(const AudioDevice: IMMDevice): integer;
    function GetSessionPID(Index: integer; const AudioDevice: IMMDevice): DWORD;
    procedure SetMasterMute(Value: boolean);
    procedure SetSessionMute(Value: boolean);
    procedure SetSessionVolume(Value: single);
    procedure SetMasterVolume(Value: single);
    property DeviceMixFormat: TWAVEFORMATEX read FMixFormat;
    property MasterChannels: DWORD read FChannels;
    property Playing: boolean read FPlaying write SetPlaying;
    property AudioDevice: TStrings read GetAudioDevice write SetAudioDevice;
    property Device: IMMDevice read GetDevice write SetDevice;
    property Icon: TIcon read GetIcon write SetIcon;
  published
    property MasterBalance: boolean read FMasterBalance write SetMasterBalance;
    property SessionVolume: boolean read FSessionVolume write SetSessionComponents;
    property SessionBalance: boolean read FSessionBalance write SetSessionBalance;
    property UseSessionInactive: boolean read FSessionInactive write SetSessionInactive;
    property Ready: boolean read FReady; // true if it's ready to operate
    property WindowXPMode: boolean read FMixerAPIReady;
    property AudioDeviceCount: Cardinal read FDeviceCount;
    property DefaultAudioDevice: TDeviceInfo read FDefaultDevice;
    property SuppressEvent: boolean read FSuppressEvent write SetSuppressEvent;
    property OnSessionStateEvent: TOnSessionStateEvent read FOnSessionStateEvent write
      FOnSessionStateEvent;
    property OnSessionCreated: TOnSessionCreated read FOnSessionCreated write FOnSessionCreated;
    property OnSessionDisconnected: TOnSessionDisconnected read FOnSessionDisconnected write
      FOnSessionDisconnected;
    property OnDeviceStateChange: TOnDeviceStateChange read FOnDeviceStateChange write
      FOnDeviceStateChange;
    property OnDefaultDeviceChange: TOnDefaultDeviceChange read FOnDefaultDeviceChange write
      FOnDefaultDeviceChange;
    property OnMasterVolumeChanged
      : TOnMasterVolumeEvent read FOnMasterVolumeEvent write FOnMasterVolumeEvent;
    property OnSessionVolumeChanged: TOnSessionVolumeEvent read FOnSessionVolumeEvent write
      FOnSessionVolumeEvent;
  end;

function PropVariantClear(pvar: PPropVariant): HResult; stdcall; external 'ole32.dll';
function SHLoadIndirectString(pszSource, psuOutBuf: PWideChar; const ccOutBuf: uint;
  var ppvReserved: Pointer): HResult; stdcall; external 'ShlWapi.DLL';

procedure Register;

implementation

uses tlhelp32;

var
  GotChannelCount: boolean = false;
  GotSessionChannelCount: boolean = false;
  ChannelCount: uint;

const

  GUID_TAudioVolume: TGUID = '{444C70B1-5C25-4800-A37F-A11223A34037}';

  cStrNames: array [0 .. 11] of string = ('Comments', 'CompanyName', 'FileDescription',
    'FileVersion', 'InternalName', 'LegalCopyright', 'LegalTrademarks', 'OriginalFilename',
    'PrivateBuild', 'ProductName', 'ProductVersion', 'SpecialBuild');

  CyclePerSec = 25;
  REFTIMES_PER_SEC: int64 = 10000000;
  REFTIMES_PER_MILLISEC = 10000;

procedure Register;
begin
  RegisterComponents('Synth', [TAudioVolume]);
end;

procedure TAudioVolume.GetProzessList(ProcList: TStringList);
var
  hSnap: THandle;
  ProcEntry: TProcessEntry32;
  s: string;
begin
  hSnap := CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);

  try
    if (hSnap <> INVALID_HANDLE_VALUE) then
    begin
      ProcEntry.dwSize := SizeOf(ProcessEntry32);
      if (Process32First(hSnap, ProcEntry)) then
      begin
        s := ProcEntry.szExeFile;
        ProcList.Add(ExtractFileName(s));
        while Process32Next(hSnap, ProcEntry) do
        begin
          s := ProcEntry.szExeFile;
          ProcList.Add(ExtractFileName(s));
        end;
      end;
    end;
  finally
    CloseHandle(hSnap);
  end;
end;
{$IFNDEF UNICODE}

procedure StrCopyW(dest, Source: PWideChar; MaxLength: DWORD);
var
  p: pbyte;
  i, k: DWORD;
  b1, b2: byte;
begin
  p := Pointer(Source);
  i := 0;
  k := MaxLength * 2;

  repeat
    b1 := p^;
    inc(p);
    inc(i);
    b2 := p^;
    inc(p);
    inc(i);
    if (b1 = 0) and (b2 = 0) then
      Break;
  until i = k;

  CopyMemory(dest, Pointer(Source), i);
end;
{$ENDIF}

// function GetDeviceInfo retrieves the device friendly name & device id for a particular device.
function GetDeviceInfo(Device: IMMDevice): TDeviceInfo;
var
  DeviceId: PWideChar;
  propertyStore: IPropertyStore;
  friendlyName: PROPVARIANT;
  deviceNameBuf: array [0 .. 127] of WideChar;
  DeviceName: PWideChar;
  HR: HResult;
begin
  Result.DeviceName := '';
  Result.DeviceId := '';
  DeviceId := nil;

  try
    HR := Device.GetId(DeviceId);
    if HR <> S_OK then
      raise Exception.Create('Unable to get device id of a device !');
    Result.DeviceId := WideString(DeviceId);

    HR := Device.OpenPropertyStore(STGM_READ, propertyStore);
    if HR <> S_OK then
      raise Exception.Create('Unable to open device property store of a device !');

    PropVariantInit(friendlyName);
    HR := propertyStore.GetValue(PKEY_Device_FriendlyName, friendlyName);
    if HR <> S_OK then
      raise Exception.Create('Unable to retrieve friendly name for a device !');

    if friendlyName.vt = VT_LPWSTR then
    begin
      DeviceName := @deviceNameBuf[0];
{$IFDEF UNICODE}
      StrCopy(DeviceName, friendlyName.pwszVal);
{$ELSE}
      StrCopyW(DeviceName, friendlyName.pwszVal, 127);
{$ENDIF}
      Result.DeviceName := DeviceName;
    end;

  finally
    if DeviceId <> nil then
      CoTaskMemFree(DeviceId);
    PropVariantClear(@friendlyName);
  end;
end;

// function GetDeviceInfo2 retrieves the device friendly name & device id for a particular device in a device collection.
function GetDeviceInfo2(DeviceCollection_: IMMDeviceCollection; DeviceIndex: LongWord): TDeviceInfo;
var
  Device: IMMDevice;
  HR: HResult;
begin
  Result.DeviceName := '';
  Result.DeviceId := '';

  try
    HR := DeviceCollection_.Item(DeviceIndex, Device);
    if HR <> S_OK then
      raise Exception.Create('Unable to get IMMDevice interface for device #' + intToStr
          (DeviceIndex));

    Result := GetDeviceInfo(Device);
  finally

  end;
end;

// function GetDeviceName retrieves the device friendly name for a particular device in a device collection.
function GetDeviceName(DeviceCollection_: IMMDeviceCollection; DeviceIndex: LongWord): WideString;
var
  Device: IMMDevice;
  DeviceId: PWideChar;
  propertyStore: IPropertyStore;
  friendlyName: PROPVARIANT;
  deviceNameBuf: array [0 .. 127] of WideChar;
  DeviceName: PWideChar;
  HR: HResult;
begin
  Result := '';
  DeviceId := nil;

  try
    HR := DeviceCollection_.Item(DeviceIndex, Device);
    if HR <> S_OK then
      raise Exception.Create('Unable to get IMMDevice interface for device #' + intToStr
          (DeviceIndex));

    HR := Device.GetId(DeviceId);
    if HR <> S_OK then
      raise Exception.Create('Unable to get device id for device #' + intToStr(DeviceIndex));

    HR := Device.OpenPropertyStore(STGM_READ, propertyStore);
    if HR <> S_OK then
      raise Exception.Create('Unable to open device property store for device #' + intToStr
          (DeviceIndex));

    PropVariantInit(friendlyName);
    HR := propertyStore.GetValue(PKEY_Device_FriendlyName, friendlyName);
    if HR <> S_OK then
      raise Exception.Create('Unable to retrieve friendly name for device #' + intToStr(DeviceIndex)
        );

    if friendlyName.vt = VT_LPWSTR then
    begin
      DeviceName := @deviceNameBuf[0];
{$IFDEF UNICODE}
      StrCopy(DeviceName, friendlyName.pwszVal);
{$ELSE}
      StrCopyW(DeviceName, friendlyName.pwszVal, 127);
{$ENDIF}
      Result := DeviceName;
    end;

  finally
    if DeviceId <> nil then
      CoTaskMemFree(DeviceId);
    PropVariantClear(@friendlyName);
  end;
end;

// ** function SetupPeakMeter should be called after initialization of application.
function TAudioVolume.SetupPeakMeter(CurProcId: DWord): boolean;
var
  HR: HResult;
  SessionCount: integer;
  SessionControl2: IAudioSessionControl2;
  i: integer;
  {CurProcId,} ProcId: Cardinal;
begin
  Result := false;

  // Session peak meter is not supported if the operating system is Windows Vista or older
  if (Win32MajorVersion < 6) or ((Win32MajorVersion = 6) and (Win32MinorVersion = 0)) then
    Exception.Create
      ('Error : Session peak meter is not supported at Windows Vista or older system');

  PeakMeter := nil;
  if SessionEnumerator = nil then
    exit;

  try
    // iterate over all the sessions
    SessionCount := 0;
    HR := SessionEnumerator.GetCount(SessionCount);
    if HR <> S_OK then
      raise Exception.Create('Error : Unable to get the number of Audio Sessions');

    // ActiveSessions := 0;
    //CurProcId := GetCurrentProcessId;
    for i := 0 to SessionCount - 1 do
    begin
      // Get Audio Session
      HR := SessionEnumerator.GetSession(i, SessionControl);
      if HR <> S_OK then
        raise Exception.Create('Error : Unable to get the Audio Session');

      // Get Audio Session Control2
      HR := SessionControl.QueryInterface(IID_IAudioSessionControl2, SessionControl2);
      if HR <> S_OK then
        raise Exception.Create('Error : Unable to get the Session Control2');

//      HR := SessionControl2.IsSystemSoundsSession;
//      if HR = S_OK then
//        continue;

      HR := SessionControl2.GetProcessID(ProcId);

      if (ProcId = CurProcId) or (CurProcId = 0) and (HR = S_OK) then
      begin
        // Get Audio Meter Information
        HR := SessionControl.QueryInterface(IID_IAudioMeterInformation, IUnknown(PeakMeter));
        if HR <> S_OK then
          raise Exception.Create('Error : Unable to get the Audio Meter Information');

        Result := true;
        Break;
      end;
    end;

  finally

  end;
end;

function TAudioVolume.Split(const Source, Delimiter: string): TSplitStrArray;
var
  spCount: integer;
  spPos: integer;
  spLength: integer;
  sTemp: string;
  aSplit: TSplitStrArray;
begin

  sTemp := Source;
  spCount := 0;
  spLength := Length(Delimiter) - 1;
  repeat
    spPos := Pos(Delimiter, sTemp);
    if spPos = 0 then
      Break
    else
    begin
      inc(spCount);
      SetLength(aSplit, spCount);
      aSplit[spCount - 1] := Copy(sTemp, 1, spPos - 1);
      Delete(sTemp, 1, spPos + spLength);
    end;

  until false;
  if Length(sTemp) > 0 then
  begin
    inc(spCount);
    SetLength(aSplit, spCount);
    aSplit[spCount - 1] := sTemp;
  end;

  Result := aSplit;
end;

function TAudioVolume.SetupDefaultAudioEndpoint(Device_Enumerator: IMMDeviceEnumerator): boolean;
var
  HR: HResult;
  MDevice: IMMDevice;
  hnsRequestedDuration: REFERENCE_TIME;
  pwfx: PWAVEFORMATEX;

begin
  Result := false;
  Device := nil;

  try
    // Get default audio endpoint device with the specified data-flow direction (eRender or eCapture)
    //  and device role.
    HR := Device_Enumerator.GetDefaultAudioEndpoint(eRender, eMultimedia, MDevice);
    if HR <> S_OK then
      raise Exception.Create('Error : Unable to get DefaultAudioEndpoint Device');

    FDefaultDevice := GetDeviceInfo(MDevice);
    Device := MDevice;

    // Get the session manager for the endpoint device.
    HR := Device.Activate(IID_IAudioSessionManager, CLSCTX_INPROC_SERVER, nil, IUnknown
        (SessionManager));
    if HR <> S_OK then
      raise Exception.Create('Error : Unable to get AudioSessionManager Interface');

    // **** Get the IChannelAudioVolume interface :
    HR := Device.Activate(IID_IAudioClient, CLSCTX_ALL, nil, IUnknown(AudioClient));
    if HR <> S_OK then
      raise Exception.Create('Error : Unable to get AudioClient Interface');

    hnsRequestedDuration := REFTIMES_PER_SEC;
    HR := AudioClient.GetMixFormat(pwfx);
    if HR <> S_OK then
      raise Exception.Create(
        'Error : Unable to retrieves the stream format that the audio engine uses');

    HR := AudioClient.Initialize(AUDCLNT_SHAREMODE_SHARED, 0, hnsRequestedDuration, 0, pwfx, nil);
    if HR <> S_OK then
      raise Exception.Create('Error : Unable to initialize AudioClient Interface');

    HR := AudioClient.GetService(IID_IChannelAudioVolume, IUnknown(ChannelAudioVolume));
    if HR <> S_OK then
      raise Exception.Create('Error : Unable to get ChannelAudioVolume Interface');

    // get SimpleAudioVolume
    HR := SessionManager.GetSimpleAudioVolume(nil, 0, SimpleAudioVolume);
    if HR <> S_OK then
      raise Exception.Create('Error : Unable to get SimpleAudioVolume Interface');

    // Get the control interface for the process-specific audio session.
    HR := SessionManager.GetAudioSessionControl(nil, false, SessionControl);
    if HR <> S_OK then
      raise Exception.Create('Error : Unable to get AudioSessionControl');

    // Register IAudioSessionEvents interface to receive notifications when session events occur.
    HR := SessionControl.RegisterAudioSessionNotification(SessionEvents);
    if HR <> S_OK then
      raise Exception.Create('Error : Unable to Register Session Notification Event');

    // We can get session enumerator if we use Windows 7 or newer system.
    if (Win32MajorVersion > 6) or ((Win32MajorVersion = 6) and (Win32MinorVersion >= 1)) then
    begin
      // Get the session manager2 for the endpoint device.
      HR := Device.Activate(IID_IAudioSessionManager2, CLSCTX_ALL, nil, IUnknown(SessionManager2));
      if HR <> S_OK then
        raise Exception.Create('Error : Unable to get AudioSessionManager2 Interface');

      // Register TAudioSessionNotification to receive a notification when a session is created.
      HR := SessionManager2.RegisterSessionNotification(SessionNotification);
      if HR <> S_OK then
        raise Exception.Create('Error : Unable to Register Session Notification');

      // Get the session enumerator
      HR := SessionManager2.GetSessionEnumerator(SessionEnumerator);
      if HR <> S_OK then
        raise Exception.Create('Error : Unable to get SessionEnumerator Interface');
    end;

    Result := true;
  finally

  end;
end;

procedure TAudioVolume.FillOutputLine;
var
  N: Cardinal;
begin
  FDeviceCount := 0;
  SetLength(DestLineInfo, 1);
  for N := 0 to mixerGetNumDevs - 1 do // mixerGetNumDevs : the number of mixer installed
  begin
    // Get Mixer line info for Speaker output
    if GetComponentControl(N, MIXERLINE_COMPONENTTYPE_DST_SPEAKERS, DestLineInfo[FDeviceCount])
      = MMSYSERR_NOERROR then
    begin
      inc(FDeviceCount);
      SetLength(FDeviceList, FDeviceCount);
      SetLength(DestLineInfo, FDeviceCount + 1);
      FDeviceList[FDeviceCount - 1].DeviceName := DestLineInfo[FDeviceCount - 1].LineName;
      FDeviceList[FDeviceCount - 1].DeviceId := intToStr(N * 10 + 1);
    end;
    // Get Mixer line info for Headphone output
    if GetComponentControl(N, MIXERLINE_COMPONENTTYPE_DST_HEADPHONES, DestLineInfo[FDeviceCount])
      = MMSYSERR_NOERROR then
    begin
      inc(FDeviceCount);
      SetLength(FDeviceList, FDeviceCount);
      SetLength(DestLineInfo, FDeviceCount + 1);
      FDeviceList[FDeviceCount - 1].DeviceName := DestLineInfo[FDeviceCount - 1].LineName;
      FDeviceList[FDeviceCount - 1].DeviceId := intToStr(N * 10 + 2);
    end;
    // Get Mixer line info for Digital output
    if GetComponentControl(N, MIXERLINE_COMPONENTTYPE_DST_DIGITAL, DestLineInfo[FDeviceCount])
      = MMSYSERR_NOERROR then
    begin
      inc(FDeviceCount);
      SetLength(FDeviceList, FDeviceCount);
      SetLength(DestLineInfo, FDeviceCount + 1);
      FDeviceList[FDeviceCount - 1].DeviceName := DestLineInfo[FDeviceCount - 1].LineName;
      FDeviceList[FDeviceCount - 1].DeviceId := intToStr(N * 10 + 3);
    end;
  end;
end;

function TAudioVolume.RefreshDeviceList(Device_Enumerator: IMMDeviceEnumerator): boolean;
var
  HR: HResult;
  N: Cardinal;

begin
  Result := false;

  if DeviceCollection <> nil then
    DeviceCollection := nil;

  try
    // Get the device info of audio endpoint devices.
    HR := Device_Enumerator.EnumAudioEndpoints(eRender, DEVICE_STATE_ACTIVE, DeviceCollection);
    if HR <> S_OK then
      raise Exception.Create('Error : Unable to retrieve device collection');

    HR := DeviceCollection.GetCount(FDeviceCount);
    if HR <> S_OK then
      raise Exception.Create('Error : Unable to get device collection length');

    if FDeviceCount > 0 then
    begin
      SetLength(FDeviceList, FDeviceCount);
      for N := 0 to (FDeviceCount - 1) do
        FDeviceList[N] := GetDeviceInfo2(DeviceCollection, N);
    end
    else
      raise Exception.Create('Error : No device to be controlled');

    Result := true;
  finally

  end;
end;

procedure TAudioVolume.ckMasterMuteClick(Sender: TObject);
begin
  if ckMasterMute.Checked <> IsMasterMuted then
    SetMasterMute(ckMasterMute.Checked);
end;

procedure TAudioVolume.ckSessionMuteClick(Sender: TObject);
begin
  if ckSessionMute.Checked <> IsMuted then
    SetSessionMute(ckSessionMute.Checked);
end;

constructor TAudioVolume.Create(AOwner: TComponent);
var
  HR: HResult;
  N: Cardinal;
  MyRegistry: TRegistry;
  ISessionEvents: TAudioSessionEvents;
  ISessionNotification: TAudioSessionNotification;
  IAudioEndpointEvents: TAudioEndpointEvents;

begin
  inherited Create(AOwner);

  DeviceEnumerator := nil;
  FMixerObj := 0;

  if Win32MajorVersion < 6 then // for Windows XP
  begin
    FillOutputLine;
    // SelectAudioDevice(0);  // * for device selection at creation.
    if not(csDesigning in ComponentState) then
      FEventHandle := AllocateHWnd(ProcessMsg);

    MyRegistry := TRegistry.Create;
    MyRegistry.RootKey := HKEY_CURRENT_USER;
    if MyRegistry.KeyExists('Software\Microsoft\Multimedia\Sound Mapper') then
      if MyRegistry.OpenKey('Software\Microsoft\Multimedia\Sound Mapper', true) then
        FDefaultDevice.DeviceName := MyRegistry.ReadString('Playback');
    MyRegistry.free;
    FDefaultDevice.DeviceId := ''; // ** Extra code shold be added to get correct device id on Windows XP
    WINMM_DEVICECHANGE := RegisterWindowMessage('winmm_devicechange');

    FMixerAPIReady := true;
    SetMasterComponents;
    SelectAudioDevice(0);

    tbMasterVolume.OnChange := tbMasterVolumeChange;
    ckMasterMute.OnClick := ckMasterMuteClick;

    FReady := true;
  end
  else
    try // Use Core Audio API if the operating system is Windows Vista or newer.
      if not(csDesigning in ComponentState) then
        FEventHandle := AllocateHWnd(ProcessMsg);
      // SessionEvents
      ISessionEvents := TAudioSessionEvents.Create;
      ISessionEvents.SetMsgHandle(FEventHandle);
      SessionEvents := ISessionEvents;
      // SessionNotification
      ISessionNotification := TAudioSessionNotification.Create;
      ISessionNotification.SetMsgHandle(FEventHandle);
      SessionNotification := ISessionNotification;
      // AudioEndpointEvents
      IAudioEndpointEvents := TAudioEndpointEvents.Create;
      IAudioEndpointEvents.SetMsgHandle(FEventHandle);
      AudioEndpointEvents := IAudioEndpointEvents;

      // Get the enumerator for the audio endpoint devices on this system.
      HR := CoCreateInstance(CLASS_IMMDeviceEnumerator, nil, CLSCTX_INPROC_SERVER,
        IID_IMMDeviceEnumerator, DeviceEnumerator);
      if HR <> S_OK then
        raise Exception.Create('Error : Unable to instantiate device enumerator');

      // Get the names of audio endpoint devices.
      HR := DeviceEnumerator.EnumAudioEndpoints(eAll, DEVICE_STATE_ACTIVE, DeviceCollection);
      if HR <> S_OK then
        raise Exception.Create('Error : Unable to retrieve device collection');

      HR := DeviceCollection.GetCount(FDeviceCount);
      if HR <> S_OK then
        raise Exception.Create('Error : Unable to get device collection length');

      if FDeviceCount > 0 then
      begin
        SetLength(FDeviceList, FDeviceCount);
        for N := 0 to (FDeviceCount - 1) do
          FDeviceList[N] := GetDeviceInfo2(DeviceCollection, N);
      end
      else
        raise Exception.Create('Error : No device to be controlled');

      if SetupDefaultAudioEndpoint(DeviceEnumerator) then
        FReady := true;

      MMNotificationClient := TMMNotificationClient.Create;
      MMNotificationClient.SetMsgHandle(FEventHandle);
      HR := DeviceEnumerator.RegisterEndpointNotificationCallback(MMNotificationClient);

      if HR <> S_OK then
        raise Exception.Create('Error : Unable to register EndpointNotificationCallback');
    finally
      SetMasterComponents;
      SelectAudioDevice(0);

      tbMasterVolume.OnChange := tbMasterVolumeChange;
      ckMasterMute.OnClick := ckMasterMuteClick;

      if not WindowXPMode then
        Timer1.Enabled := true;
    end;
end;

destructor TAudioVolume.Destroy;
begin
  if MMNotificationClient <> nil then
  begin
    DeviceEnumerator.UnregisterEndpointNotificationCallback(MMNotificationClient);
    MMNotificationClient.free;
  end;

  if AudioEndpointVolume <> nil then
    AudioEndpointVolume.UnregisterControlChangeNotify(AudioEndpointEvents);

  if SessionControl <> nil then
    SessionControl.UnregisterAudioSessionNotification(SessionEvents);

  if SessionManager2 <> nil then
    SessionManager2.UnregisterSessionNotification(SessionNotification);

  if not(csDesigning in ComponentState) then
  begin
    if HMixer <> 0 then
      mixerClose(HMixer);
    DeallocateHWnd(FEventHandle);
  end;

  SetLength(FDeviceList, 0);
  SetLength(DestLineInfo, 0);

  if Assigned(FAudioSessionList.Icon) then
    FreeAndNil(FAudioSessionList.Icon);

  FAudioDevice.free;

  inherited;
end;

function TAudioVolume.SelectAudioDevice(DeviceIndex: Cardinal): boolean;
var
  HR: HResult;
  Device: IMMDevice;
  pDeviceId: PWideChar;
  AudioClient: IAudioClient;
  pFormat: PWAVEFORMATEX;
  Items: TStringList;

begin
  Result := false;

  if FDeviceCount = 0 then
    exit;
  if DeviceIndex > (FDeviceCount - 1) then
    exit;

  if FDeviceList[DeviceIndex].DeviceId = FSelectedDevice.DeviceId then
    exit;

  if Win32MajorVersion < 6 then
  begin
    if HMixer <> 0 then
    begin
      mixerClose(HMixer);
      HMixer := 0;
    end;

    FSelectedDevice.DeviceName := FDeviceList[DeviceIndex].DeviceName;
    FSelectedDevice.DeviceId := FDeviceList[DeviceIndex].DeviceId;
    FMixerObj := DestLineInfo[DeviceIndex].MixerId;
    FChannels := DestLineInfo[DeviceIndex].Channels;
    MasterMute := DestLineInfo[DeviceIndex].MuteControl;
    MasterVol := DestLineInfo[DeviceIndex].VolumeControl;

    if mixerOpen(@HMixer, FMixerObj, FEventHandle, 0, CALLBACK_WINDOW + MIXER_OBJECTF_HMIXER)
      = MMSYSERR_NOERROR then
      Result := true;

    Items := TStringList.Create;
    tbMasterVolume.Position := round((1.0 - GetMasterVolume) * tbMasterVolume.Max);
    ckMasterMute.Checked := IsMasterMuted;
{$IFDEF UNICODE}
    ExtractStrings(['('], [], PWideChar(WideString(FSelectedDevice.DeviceName)), Items);
{$ELSE}
    ExtractStrings(['('], [], PChar(AnsiString(FSelectedDevice.DeviceName)), Items);
{$ENDIF}
    lblMasterCaption.Caption := Items[0];
    lblMasterCaption.Left := 5;
    lblMasterCaption.Top := 40;
    lblMasterCaption.Width := FMGroupBox.Width + 5;
    lblMasterCaption.Alignment := taCenter;
    lblMasterCaption.AutoSize := true;
    Items.free;
    exit;
  end;

  try
    if AudioEndpointVolume <> nil then
    begin
      HR := AudioEndpointVolume.UnregisterControlChangeNotify(AudioEndpointEvents);
      if HR <> S_OK then
        raise Exception.Create('Unable to unregister AudioEndpointEvents for device #' + intToStr
            (DeviceIndex));

      AudioEndpointVolume := nil; // AudioEndpointVolume is released by automatic action
      MasterMeter := nil; // MasterMeter is released by automatic action
    end;

    HR := DeviceCollection.Item(DeviceIndex, Device);
    if HR <> S_OK then
      raise Exception.Create('Unable to get IMMDevice interface for device #' + intToStr
          (DeviceIndex));

    HR := Device.Activate(IID_IAudioClient, CLSCTX_INPROC_SERVER, nil, IUnknown(AudioClient));
    if HR <> S_OK then
      raise Exception.Create('Unable to get AudioClient interface for device #' + intToStr
          (DeviceIndex));

    HR := AudioClient.GetMixFormat(pFormat); // Get mixer format
    if HR <> S_OK then
      raise Exception.Create('Unable to get MixFormat for device #' + intToStr(DeviceIndex));

    FMixFormat := pFormat^;
    FChannels := FMixFormat.nChannels;
    CoTaskMemFree(pFormat);

    HR := Device.Activate(IID_IAudioEndpointVolume, CLSCTX_INPROC_SERVER, nil, IUnknown
        (AudioEndpointVolume));
    if HR <> S_OK then
      raise Exception.Create('Unable to get IAudioEndpointVolume interface for device #' + intToStr
          (DeviceIndex));

    // Get IAudioMeterInformation interface to use for peak detect
    HR := Device.Activate(IID_IAudioMeterInformation, CLSCTX_ALL, nil, IUnknown(MasterMeter));
    if HR <> S_OK then
      raise Exception.Create('Unable to get MasterMeter interface for device #' + intToStr
          (DeviceIndex));

    HR := AudioEndpointVolume.RegisterControlChangeNotify(AudioEndpointEvents);
    if HR <> S_OK then
      raise Exception.Create('Unable to register AudioEndpointEvents for device #' + intToStr
          (DeviceIndex));

    FSelectedDevice.DeviceName := GetDeviceName(DeviceCollection, DeviceIndex);
    Device.GetId(pDeviceId);
    FSelectedDevice.DeviceId := WideString(pDeviceId);
    Result := true;
  finally
    Items := TStringList.Create;

    tbMasterVolume.Position := round((1.0 - GetMasterVolume) * tbMasterVolume.Max);
    ckMasterMute.Checked := IsMasterMuted;
{$IFDEF UNICODE}
    ExtractStrings(['('], [], PWideChar(WideString(FSelectedDevice.DeviceName)), Items);
{$ELSE}
    ExtractStrings(['('], [], PChar(AnsiString(FSelectedDevice.DeviceName)), Items);
{$ENDIF}
    lblMasterCaption.Caption := Items[0];
    lblMasterCaption.Left := 5;
    lblMasterCaption.Top := 40;
    lblMasterCaption.Width := FMGroupBox.Width + 5;
    lblMasterCaption.Alignment := taCenter;
    lblMasterCaption.AutoSize := true;

    Items.free;
  end;
end;

procedure TAudioVolume.ProcessMsg(var Msg: TMessage);
var
  PVolMuteRec: PVolMute;
  Volume: single;
  Muted: boolean;
  Device: IMMDevice;
  DeviceId: WideString;
  DeviceInfo: TDeviceInfo;
  HR: HResult;
  NewState: DWORD;
  MyRegistry: TRegistry;
begin
  case Msg.Msg of

    WM_EndpointVolume:
      begin
        PVolMuteRec := PVolMute(Msg.WPARAM);
        if Assigned(tbMasterVolume) then
        begin
          tbMasterVolume.Position := round((1.0 - PVolMuteRec^.Volume) * tbMasterVolume.Max);
          ckMasterMute.Checked := PVolMuteRec^.Muted;
        end;
        if Assigned(tbMasterBalance) then
          AdjustMasterBalancePos;

        if Assigned(FOnMasterVolumeEvent) then
          FOnMasterVolumeEvent(MyVolSet, PVolMuteRec^.Volume, PVolMuteRec^.Muted);
      end;

    WM_VolumeEvent:
      begin
        PVolMuteRec := PVolMute(Msg.WPARAM);

        if MySessionVolSet or Assigned(FOnSessionVolumeEvent) then
        begin
          if Assigned(tbSessionVolume) or Assigned(FOnSessionVolumeEvent) then
          begin
            tbSessionVolume.Position := round((1.0 - PVolMuteRec^.Volume) * tbSessionVolume.Max);
            ckSessionMute.Checked := PVolMuteRec^.Muted;
          end;

          if (MySessionVolSet and (abs(PVolMuteRec^.Volume - MyVolVal) < 0.0001)) then
            MySessionVolSet := false
          else if (MyMuteSet and (PVolMuteRec^.Muted = MyMuteVal)) then
            MyMuteSet := false;

          if Assigned(tbSessionBalance) then
            AdjustSessionBalancePos;
        end;
      end;

    WM_SessionStateEvent:
      begin
        if Assigned(FOnSessionStateEvent) then
          FOnSessionStateEvent(self, Msg.WPARAM);
      end;

    WM_DeviceStateChange, WM_DeviceAdded, WM_DeviceRemoved:
      begin
        DeviceId := WideString(PWideChar(Msg.WPARAM));
        NewState := Msg.LPARAM;
        // Just refresh device list to simplify the logic.
        RefreshDeviceList(DeviceEnumerator);
        //  SetupDefaultAudioEndpoint(DeviceEnumerator);
        if Assigned(OnDeviceStateChange) then
        begin
          if Msg.Msg = WM_DeviceStateChange then
            FOnDeviceStateChange(DeviceId, NewState)
          else if Msg.Msg = WM_DeviceAdded then
            FOnDeviceStateChange(DeviceId, DEVICE_STATE_ADDED)
          else if Msg.Msg = WM_DeviceRemoved then
            FOnDeviceStateChange(DeviceId, DEVICE_STATE_REMOVED);
        end;
      end;

    WM_SessionDisconnected:
      begin
        FAudioSessionList.SessionDisconnected := cAudioSessionDisconnected
          [TAudioSessionDisconnected(Msg.WPARAM)];
        if Assigned(FOnSessionDisconnected) then
          FOnSessionDisconnected(self, Msg.WPARAM);
      end;

    WM_SessionCreate:
      begin
        if Assigned(FOnSessionCreated) then
          FOnSessionCreated(IAudioSessionControl(@Msg.WPARAM));
      end;

    WM_DefaultDeviceChange:
      begin
        HR := DeviceEnumerator.GetDefaultAudioEndpoint(eRender, eMultimedia, Device);
        if HR <> S_OK then
          raise Exception.Create('Error : Unable to get DefaultAudioEndpoint Device');
        DeviceInfo := GetDeviceInfo(Device);
        if DeviceInfo.DeviceId <> FDefaultDevice.DeviceId then
        begin
          FDefaultDevice := DeviceInfo;
          if Assigned(OnDefaultDeviceChange) then
            FOnDefaultDeviceChange(DeviceInfo);
        end;
      end;

    MM_MIXM_LINE_CHANGE:
      begin
        // wParam = (WPARAM) hMixer lParam = (LPARAM) dwLineID

      end;

    MM_MIXM_CONTROL_CHANGE:
      begin
        // wParam = (WPARAM) hMixer    lParam = (LPARAM) dwControlID
        if (DWORD(Msg.LPARAM) <> MasterMute.dwControlID) and
          (DWORD(Msg.LPARAM) <> MasterVol.dwControlID) then
          exit;

        Muted := IsMasterMuted;
        Volume := GetMasterVolume;
        if (DWORD(Msg.LPARAM) = MasterMute.dwControlID) then
        begin
          if (MyMuteSet and (Muted = MyMuteVal)) then
            MyMuteSet := false;
        end
        else
        // for (Msg.LParam = MasterVol.dwControlID)
          if (MyVolSet and (abs(Volume - MyVolVal) < 0.0001)) then
          MyVolSet := false;

        if Assigned(tbMasterVolume) then
        begin
          tbMasterVolume.Position := round((1.0 - Volume) * tbMasterVolume.Max);
          ckMasterMute.Checked := Muted;
        end;

        if Assigned(tbMasterBalance) then
          AdjustMasterBalancePos;
      end;
  else
    if Msg.Msg = WINMM_DEVICECHANGE then
    begin
      MyRegistry := TRegistry.Create;
      MyRegistry.RootKey := HKEY_CURRENT_USER;
      if MyRegistry.KeyExists('Software\Microsoft\Multimedia\Sound Mapper') then
        if MyRegistry.OpenKey('Software\Microsoft\Multimedia\Sound Mapper', true) then
          FDefaultDevice.DeviceName := MyRegistry.ReadString('Playback');
      MyRegistry.free;
    end
    else
      Msg.Result := DefWindowProc(FEventHandle, Msg.Msg, Msg.WPARAM, Msg.LPARAM);
  end;
end;

procedure TAudioVolume.SetSuppressEvent(Value: boolean);
var
  HR: HResult;
begin
  if Value = FSuppressEvent then
    exit;
  if (SessionControl = nil) and (AudioEndpointVolume = nil) then
    exit;

  if Win32MajorVersion < 6 then
  begin
    FSuppressEvent := Value;
    exit;
  end;

  try
    if AudioEndpointVolume <> nil then
    begin
      if Value then
        HR := AudioEndpointVolume.UnregisterControlChangeNotify(AudioEndpointEvents)
      else
        HR := AudioEndpointVolume.RegisterControlChangeNotify(AudioEndpointEvents);
      if HR <> S_OK then
        raise Exception.Create('Unable to (un)register AudioEndpointEvents');
    end;

    if (SessionControl <> nil) then
    begin
      if Value then
        HR := SessionControl.UnregisterAudioSessionNotification(SessionEvents)
      else
        HR := SessionControl.RegisterAudioSessionNotification(SessionEvents);
      if HR <> S_OK then
        raise Exception.Create('Unable to (un)register AudioSessionNotification');
    end;

    if (SessionManager2 <> nil) then
    begin
      if Value then
        HR := SessionManager2.UnregisterSessionNotification(SessionNotification)
      else
        HR := SessionManager2.RegisterSessionNotification(SessionNotification);
      if HR <> S_OK then
        raise Exception.Create('Unable to (un)register SessionNotification');
    end;

  finally

  end;

  FSuppressEvent := Value;
end;
{$EXTERNALSYM SHGetFileInfoW}
function SHGetFileInfoW(pszPath: PWideChar; dwFileAttributes: DWORD; var psfi: TSHFileInfoW;
  cbFileInfo, uFlags: uint): DWORD; stdcall; external shell32 name 'SHGetFileInfoW';

function TAudioVolume.SetSysItemIcon(wsItems: PWideChar): TIcon;
var
  fi: TSHFileInfoW;
begin
  Result := nil;

  ZeroMemory(@fi, SizeOf(TSHFileInfo));
  if SHGetFileInfoW(PWideChar(wsItems), 0, fi, SizeOf(TSHFileInfo),
    SHGFI_DISPLAYNAME Or SHGFI_SYSICONINDEX or SHGFI_LARGEICON or SHGFI_ICON) <> 0 then
  begin
    Icon := TIcon.Create;
    Icon.Handle := fi.HICON;
    Result := Icon;
  end;

end;

procedure TAudioVolume.AdjustMasterBalancePos;
var
  HR: HResult;
  Channels: uint;
  leftVol, rightVol: single;
  BalanceVal: DWORD;
  Volume: TVolume;
begin

  if not Assigned(tbMasterBalance) then
    exit;

  if Win32MajorVersion < 6 then
  begin
    if FMixerAPIReady then
    begin // **** Pre-Check if balance controls is available
      if (FChannels < 2) or ((MasterVol.fdwControl and MIXERCONTROL_CONTROLF_UNIFORM)
          = MIXERCONTROL_CONTROLF_UNIFORM) then
        exit;

      if GetMixerVolume(FMixerObj, FChannels, MasterVol, Volume) = MMSYSERR_NOERROR then
      begin
        leftVol := Volume[0] / MasterVol.Bounds.dwMaximum;
        rightVol := Volume[1] / MasterVol.Bounds.dwMaximum;

        if leftVol > rightVol then
          BalanceVal := round((rightVol / leftVol) * MasterBalance_Mid)
        else if leftVol < rightVol then
          BalanceVal := round((2.0 - (leftVol / rightVol)) * MasterBalance_Mid)
        else
          BalanceVal := MasterBalance_Mid;

        tbMasterBalance.OnChange := nil;

        tbMasterBalance.Position := BalanceVal;
        tbMasterBalance.OnChange := tbMasterBalanceChange;
      end;
      exit;
    end;
  end;

  if AudioEndpointVolume = nil then
    exit;

  // following code to set Balance at intializing is not effective (always get leftVol = rightVol = 1)
  HR := AudioEndpointVolume.GetChannelCount(Channels);
  if HR <> S_OK then
    tbMasterBalance.Position := MasterBalance_Mid
  else if Channels < 2 then
    tbMasterBalance.Position := MasterBalance_Mid
  else
  begin
    // Get the current Volume of Left Channel & Right channel
    AudioEndpointVolume.GetChannelVolumeLevelScalar(0, leftVol);
    AudioEndpointVolume.GetChannelVolumeLevelScalar(1, rightVol);

    // Get the Balance value according to the volume of left & right channel
    if leftVol = rightVol then
      BalanceVal := MasterBalance_Mid
    else if leftVol > rightVol then
      BalanceVal := round((rightVol / leftVol) * MasterBalance_Mid)
    else
      BalanceVal := round((2.0 - (leftVol / rightVol)) * MasterBalance_Mid);

    tbMasterBalance.OnChange := nil;

    tbMasterBalance.Position := BalanceVal;
    tbMasterBalance.OnChange := tbMasterBalanceChange;
  end;
end;

procedure TAudioVolume.AdjustSessionBalancePos;
var
  HR: HResult;
  Channels: uint;
  leftVol, rightVol: single;
  BalanceVal: DWORD;

begin

  if not Assigned(tbSessionBalance) then
    exit;

  if AudioEndpointVolume = nil then
    exit;

  // following code to set Balance at intializing is not effective (always get leftVol = rightVol = 1)
  HR := AudioEndpointVolume.GetChannelCount(Channels);
  if HR <> S_OK then
    tbSessionBalance.Position := SessionBalance_Mid
  else if Channels < 2 then
    tbSessionBalance.Position := SessionBalance_Mid
  else
  begin
    // Get the current Volume of Left Channel & Right channel
    AudioEndpointVolume.GetChannelVolumeLevelScalar(0, leftVol);
    AudioEndpointVolume.GetChannelVolumeLevelScalar(1, rightVol);

    // Get the Balance value according to the volume of left & right channel
    if leftVol = rightVol then
      BalanceVal := SessionBalance_Mid
    else if leftVol > rightVol then
      BalanceVal := round((rightVol / leftVol) * SessionBalance_Mid)
    else
      BalanceVal := round((2.0 - (leftVol / rightVol)) * SessionBalance_Mid);

    tbSessionBalance.OnChange := nil;

    tbSessionBalance.Position := BalanceVal;
    tbSessionBalance.OnChange := tbSessionBalanceChange;
  end;

end;

function TAudioVolume.AudioDeviceInfo(DeviceIndex: Cardinal): TDeviceInfo;
begin
  Result.DeviceName := '';
  Result.DeviceId := '';

  if FDeviceCount = 0 then
    exit;
  if DeviceIndex > (FDeviceCount - 1) then
    exit;

  Result := FDeviceList[DeviceIndex];
end;

function TAudioVolume.AudioSessionInfo(SessionIndex: integer): TAudioSessionInfo;
var
  Lbl: TLabel;
  tmpStr: string;
  bNull: Bool;
  i: integer;
begin
  Result.ActiveSession := '';
  Result.DisplayName := '';
  Result.StateSession := '';
  Result.IconPath := '';
  Result.SessionIdentifier := '';
  Result.ProcessID := '';
  Result.IsSystemSoundsSession := -1;
  Result.SessionGuid := StringToGUID('{00000000-0000-0000-0000-000000000000}');
  Result.Device := nil;
  Result.PID := 0;

  bNull := false;
  if Assigned(FSGroupBox) then
  begin
    tmpStr := ChangeFileExt(FAudioSessionList.DisplayName, '');
    if tmpStr <> '' then
    begin
      Lbl := TLabel.Create(nil);
      Lbl.Caption := tmpStr;
      Lbl.AutoSize := true;
      Hint := tmpStr;
      // check for blank
      for i := 1 to Length(tmpStr) do
      begin
        if tmpStr[i] = ' ' then
          bNull := true;
      end;

      if not bNull then
      begin
        if Lbl.Width > FSGroupBox.Width - 85 then
        begin
          Lbl.Caption := Lbl.Caption + '...';
          while Lbl.Width > FSGroupBox.Width - 85 do
            Lbl.Caption := Copy(Lbl.Caption, 1, Length(Lbl.Caption) - 8) + '...';
        end;
      end;

      lblSessionCaption.WordWrap := bNull;
      lblSessionCaption.Caption := Lbl.Caption;
      tmpStr := '';
      FreeAndNil(Lbl);
    end;
  end;

  Result := FAudioSessionList;
end;

function TAudioVolume.IIF(AValue: boolean; const ATrue, AFalse: integer): integer;
begin

  if AValue then
    Result := (ATrue)
  else
    Result := (AFalse);
end;

function TAudioVolume.IsMasterMuted: boolean;
var
  b: Bool;
  HR: HResult;
begin
  Result := false;

  if Win32MajorVersion < 6 then
  begin
    if FMixerAPIReady then
      Result := IsMute(FMixerObj, MasterMute);
    exit;
  end;

  if AudioEndpointVolume = nil then
    exit;

  HR := AudioEndpointVolume.GetMute(b);
  if HR = S_OK then
    Result := b;
end;

procedure TAudioVolume.SetMasterBalance(const Value: boolean);
begin

  FMasterBalance := Value;
  if FMasterBalance and Assigned(FMGroupBox) then
  begin
    tbMasterBalance := TTrackBar.Create(self);
    tbMasterBalance.Parent := FMGroupBox.Parent;
    tbMasterBalance.Orientation := trHorizontal;
    tbMasterBalance.Max := MasterBalance_Max;
    tbMasterBalance.Min := MasterBalance_Min;
    tbMasterBalance.TickStyle := tsAuto;
    tbMasterBalance.Frequency := 3;
    AdjustMasterBalancePos;

    tbMasterBalance.OnChange := tbMasterBalanceChange;

    lblMasterBalance := TLabel.Create(self);
    lblMasterBalance.Parent := FMGroupBox.Parent;
    lblMasterBalance.Caption := 'Balance';
  end
  else
  begin
    if Assigned(tbMasterBalance) then
    begin
      FreeAndNil(tbMasterBalance);
      FreeAndNil(lblMasterBalance);
    end;
  end;
  PostMessage(Handle, WM_SIZE, 0, 0);
end;

procedure TAudioVolume.SetAudioDevice(const Value: TStrings);
begin
  FAudioDevice.Assign(Value);
end;

procedure TAudioVolume.SetMasterComponents;
begin

  FMGroupBox := TGroupBox.Create(self);
  FMGroupBox.Parent := self;

  lblMasterCaption := TLabel.Create(self);
  lblMasterCaption.Parent := FMGroupBox.Parent;
  lblMasterCaption.Alignment := taCenter;
  lblMasterCaption.AutoSize := true;

  if (csDesigning in ComponentState) then
    if Height = 0 then
      Height := 145;

  pbMasterMeter := TProgressBar.Create(self);
  pbMasterMeter.Parent := FMGroupBox.Parent;
  pbMasterMeter.Orientation := pbVertical;
  pbMasterMeter.Max := 100;
  pbMasterMeter.Min := 0;
  pbMasterMeter.Smooth := true;

  lblMasterPeak := TLabel.Create(self);
  lblMasterPeak.Parent := FMGroupBox.Parent;
  lblMasterPeak.Caption := 'M';

  pbMasterMeterL := TProgressBar.Create(self);
  pbMasterMeterL.Parent := FMGroupBox.Parent;
  pbMasterMeterL.Orientation := pbVertical;
  pbMasterMeterL.Max := 100;
  pbMasterMeterL.Min := 0;
  pbMasterMeterL.Smooth := true;

  lblMasterPeakL := TLabel.Create(self);
  lblMasterPeakL.Parent := FMGroupBox.Parent;
  lblMasterPeakL.Caption := 'L';

  tbMasterVolume := TTrackBar.Create(self);
  tbMasterVolume.Parent := FMGroupBox.Parent;
  tbMasterVolume.Orientation := trVertical;
  tbMasterVolume.Max := 32;
  tbMasterVolume.Min := 0;
  tbMasterVolume.TickStyle := tsAuto;
  tbMasterVolume.Frequency := 3;
  tbMasterVolume.Position := 0;

  pbMasterMeterR := TProgressBar.Create(self);
  pbMasterMeterR.Parent := FMGroupBox.Parent;
  pbMasterMeterR.Orientation := pbVertical;
  pbMasterMeterR.Max := 100;
  pbMasterMeterR.Min := 0;
  pbMasterMeterR.Smooth := true;

  lblMasterPeakR := TLabel.Create(self);
  lblMasterPeakR.Parent := FMGroupBox.Parent;
  lblMasterPeakR.Caption := 'R';

  ckMasterMute := TCheckbox.Create(self);
  ckMasterMute.Parent := FMGroupBox.Parent;
  ckMasterMute.Caption := 'Mute';

  Timer1 := TTimer.Create(self);
  Timer1.Interval := 30;
  Timer1.OnTimer := Timer1Timer;
  Timer1.Enabled := false;
end;

procedure TAudioVolume.SetDevice(const Value: IMMDevice);
begin

  FDevice := Value;
end;

procedure TAudioVolume.SetDiplayNameAndIcon(PID: DWORD; Index: integer;
  out SessionAudioVolume: ISimpleAudioVolume; SessionControl: IAudioSessionControl;
  SessionControl2: IAudioSessionControl2; State: uint);
var
  HR: HResult;
  pDisplayName: PWideChar;
  pIconPath: PWideChar;
  pSessionIdentifier: PWideChar;
  FileName: WideString;
begin

  HR := SessionControl.RegisterAudioSessionNotification(SessionEvents);
  if HR <> S_OK then
    raise Exception.Create('Error : Unable to Register Session Notification Event');

  FAudioSessionList.StateSession := cAudioSessionState[TAudioSessionState(State)];
  HR := SessionControl2.GetGroupingParam(@GroupGuid);
  if Succeeded(HR) then
  begin
    FAudioSessionList.SessionGuid := GroupGuid;
    HR := SessionControl2.GetDisplayName(pDisplayName);
    if HR = S_OK then
    begin
      FAudioSessionList.DisplayName := pDisplayName;
      if PID = 0 then
      begin
        FAudioSessionList.DisplayName := LoadDisplayNamefromDLL(pDisplayName);
      end
      else
      begin
        if AudioSessionInfo(Index).DisplayName = '' then
        begin
          FileName := GetProcessPathByPID(PID);
          if FileName <> '' then
          begin
            FAudioSessionList.DisplayName := GetVerInfo(FileName);
            if FAudioSessionList.DisplayName = '' then
            begin
              FAudioSessionList.DisplayName := ExtractFileName(ChangeFileExt(FileName, ''));
              CoTaskMemFree(pDisplayName);
            end;
          end;
        end;
      end;

      HR := SessionControl2.GetIconPath(pIconPath);
      if Succeeded(HR) then
      begin
        FAudioSessionList.IconPath := pIconPath;
        if PID = 0 then
        begin
          LoadIconPathfromResource(pIconPath)
        end
        else
        begin
          if FAudioSessionList.IconPath = '' then
            FAudioSessionList.IconPath := GetProcessPathByPID(PID);
          CoTaskMemFree(pIconPath);
          FAudioSessionList.Icon := SetSysItemIcon(PWideChar(GetProcessPathByPID(PID)));
        end;

        HR := SessionControl2.GetSessionIdentifier(pSessionIdentifier);
        if Succeeded(HR) then
        begin
          FAudioSessionList.SessionIdentifier := pSessionIdentifier;
          HR := SessionControl2.QueryInterface(ISimpleAudioVolume, SimpleAudioVolume);
          if Succeeded(HR) then
            SessionAudioVolume := SimpleAudioVolume;
        end;
      end;
    end;
  end;
end;

procedure TAudioVolume.SetIcon(const Value: TIcon);
begin

  FIcon := Value;
end;

procedure TAudioVolume.SetMasterMute(Value: boolean);
var
  b: Bool;
  i: integer;
  HR: HResult;
begin
  if Win32MajorVersion < 6 then
  begin
    if FMixerAPIReady then
    begin
      MyMuteSet := true;
      MyMuteVal := Value;
      MuteLine(FMixerObj, MasterMute, Value);
    end;
    exit;
  end;

  if AudioEndpointVolume = nil then
    exit;

  HR := AudioEndpointVolume.GetMute(b);
  if HR = S_OK then
    if Value <> b then
    begin
      // ** We should use very tricky code as follows.
      // ** If we use normal sentence like "AudioEndpointVolume.SetSysSoundMute(Value, nil);"
      // ** it does not operate.
      if b then
        i := 0
      else
        i := 1;
      AudioEndpointVolume.SetMute(Bool(i), @GUID_TAudioVolume);
    end;
end;

function TAudioVolume.GetMasterVolume: single;
var
  f: single;
  Volume: TVolume;
  maxVolume: DWORD;
  HR: HResult;
begin
  Result := 0;

  if Win32MajorVersion < 6 then
  begin
    if FMixerAPIReady then
      if GetMixerVolume(FMixerObj, FChannels, MasterVol, Volume) = MMSYSERR_NOERROR then
      begin
        if FChannels > 1 then
        begin
          if (MasterVol.fdwControl and MIXERCONTROL_CONTROLF_UNIFORM)
            = MIXERCONTROL_CONTROLF_UNIFORM then
            maxVolume := Volume[0]
          else
            maxVolume := Max(Volume[0], Volume[1])
        end
        else
          maxVolume := Volume[0];
        Result := maxVolume / MasterVol.Bounds.dwMaximum;
      end;
    exit;
  end;

  if AudioEndpointVolume = nil then
    exit;

  HR := AudioEndpointVolume.GetMasterVolumeLevelScalar(f);
  if HR = S_OK then
    Result := f;
end;

procedure TAudioVolume.SetMasterVolume(Value: single);
var
  i: integer;
  dwVolume: DWORD;
  dwBalance: DWORD;
  Volume: TVolume;
begin
  if Value < 0 then
    exit;
  if Value > 1.0 then
    exit;

  if Win32MajorVersion < 6 then
  begin
    if FMixerAPIReady then
    begin
      dwVolume := round(Value * MasterVol.Bounds.dwMaximum);
      for i := 0 to FChannels - 1 do
        Volume[i] := dwVolume;

      if (FChannels > 1) and ((MasterVol.fdwControl and MIXERCONTROL_CONTROLF_UNIFORM) = 0) then
      begin
        dwBalance := GetMasterBalance;
        if dwBalance < MasterBalance_Mid then // left sided balance
          Volume[1] := round(dwVolume * (dwBalance / MasterBalance_Mid))
        else if dwBalance > MasterBalance_Mid then // right sided balance
          Volume[0] := round(dwVolume * (2.0 - (dwBalance / MasterBalance_Mid)));
      end;

      MyVolSet := true;
      MyVolVal := Value;
      SetMixerVolume(FMixerObj, FChannels, MasterVol, Volume);
    end;
    exit;
  end;

  if AudioEndpointVolume = nil then
    exit;

  AudioEndpointVolume.SetMasterVolumeLevelScalar(Value, @GUID_TAudioVolume);
end;

function TAudioVolume.IsMuted: boolean;
var
  b: LongBool;
  HR: HResult;
begin
  Result := false;

  if Win32MajorVersion < 6 then
    exit;
  if SimpleAudioVolume = nil then
    exit;

  HR := SimpleAudioVolume.GetMute(b);
  if HR = S_OK then
    Result := b;
end;

function TAudioVolume.LoadDisplayNamefromDLL(DisplayName: PWideChar): WideString;
var
  Lbuf: array [0 .. 4095] of WideChar;
begin

  ZeroMemory(@Lbuf, SizeOf(Lbuf));
  if SHLoadIndirectString(PWideChar(DisplayName), Lbuf, Length(Lbuf), Pointer(nil^)) = S_OK then
    Result := WideString(Lbuf)
  else
    RaiseLastOsError;
end;

procedure TAudioVolume.LoadIconPathfromResource(IconPath: PWideChar);
var
  Module: HModule;
  FileName: WideString;
  lpDest: array [0 .. 32768] of WideChar;
  SplitID: TSplitStrArray;
  rSplitID: string;
  ID: DWORD;
begin

  // Remove First @ Letter
  FileName := MidStr(IconPath, 2, Length(IconPath));
  // Extract ID
  SplitID := Split(FileName, ',');
  rSplitID := SplitID[ High(SplitID)];
  rSplitID := MidStr(rSplitID, 2, Length(rSplitID));

  ID := StrToInt(rSplitID);
  // Remove ID from Path
  FileName := SplitID[ Low(SplitID)];

  FillChar(lpDest, SizeOf(lpDest), 0);
  if ExpandEnvironmentStrings(PWideChar(FileName), lpDest, 32767) <> 0 then
  begin
    // Load the Lib
    Module := LoadLibraryExW(lpDest, 0, LOAD_LIBRARY_AS_DATAFILE);
    // check for error
    if Module <> INVALID_HANDLE_VALUE then
    begin
      Icon := TIcon.Create;
      Icon.Handle := LoadIcon(Module, MakeIntResource(ID));
      if Module <> 0 then
        FreeLibrary(Module);
    end;
  end;
end;

procedure TAudioVolume.SetPlaying(const Value: boolean);
begin
  FPlaying := Value;
end;

function TAudioVolume.GetVerInfo(FileName: WideString): WideString;
var
  VI: TVerInfo;
  TransIdx: integer;
  Trans: string;
  StrIdx: integer;
  StrName: string;
begin
  VI := TVerInfo.Create(FileName);
  try
    if VI.HasVerInfo then
    begin
      for TransIdx := 0 to Pred(VI.TranslationCount) do
      begin
        Trans := VI.Translations[TransIdx];

        // display all standard strings
        for StrIdx := Low(cStrNames) to High(cStrNames) do
        begin
          StrName := cStrNames[StrIdx];
          if StrName = 'FileDescription' then
          begin
            Result := VI.Strings[Trans, StrName];
            Break;
          end;
        end;
      end;
    end
    else
      Result := '';
  finally
    VI.free;
  end;
end;

function TAudioVolume.GetVolume: single;
var
  f: single;
  HR: HResult;
begin
  Result := 0;
  if Win32MajorVersion < 6 then
    exit;
  if SimpleAudioVolume = nil then
    exit;

  HR := SimpleAudioVolume.GetMasterVolume(f);
  if HR = S_OK then
    Result := f;
end;

procedure TAudioVolume.tbMasterBalanceChange(Sender: TObject);
var
  Balance: DWORD;
  rightVol: single;
  leftVol: single;
  refVol: single;
  Channels: uint;
  HR: HResult;
begin
  if Win32MajorVersion < 6 then // for Windows XP
  begin
    if FMixerAPIReady then
    begin
      Balance := Min(MasterBalance_Max, tbMasterBalance.Position);

      SetMixerBalance(FMixerObj, FChannels, MasterVol, Balance, MasterBalance_Max,
        MasterBalance_Mid);
    end;
    exit;
  end;

  if AudioEndpointVolume = nil then
    exit;

  HR := AudioEndpointVolume.GetChannelCount(Channels);
  if HR = S_OK then
    if Channels < 2 then
      exit;

  // Get the current Volume of Left Channel & Right channel
  AudioEndpointVolume.GetChannelVolumeLevelScalar(0, leftVol);
  AudioEndpointVolume.GetChannelVolumeLevelScalar(1, rightVol);
  refVol := Max(leftVol, rightVol);

  Balance := Min(MasterBalance_Max, tbMasterBalance.Position);
  if Balance < MasterBalance_Mid then // left sided balance
  begin
    leftVol := refVol;
    rightVol := refVol * (Balance / MasterBalance_Mid);
  end
  else if Balance = MasterBalance_Mid then // center balance
  begin
    leftVol := refVol;
    rightVol := refVol;
  end
  else // right sided balance
  begin
    leftVol := refVol * (2.0 - (Balance / MasterBalance_Mid));
    rightVol := refVol;
  end;

  // Set the Volume of Left Channel & Right channel
  AudioEndpointVolume.SetChannelVolumeLevelScalar(0, leftVol, @GUID_TAudioVolume);
  AudioEndpointVolume.SetChannelVolumeLevelScalar(1, rightVol, @GUID_TAudioVolume);

end;

procedure TAudioVolume.tbMasterVolumeChange(Sender: TObject);
begin
  SetMasterVolume((tbMasterVolume.Max - tbMasterVolume.Position) / tbMasterVolume.Max);
end;

procedure TAudioVolume.tbSessionBalanceChange(Sender: TObject);
var
  Balance: DWORD;
  rightVol: single;
  leftVol: single;
  refVol: single;
  Channels: uint;
  HR: HResult;
begin
  if AudioEndpointVolume = nil then
    exit;

  HR := AudioEndpointVolume.GetChannelCount(Channels);
  if HR = S_OK then
    if Channels < 2 then
      exit;

  // Get the current Volume of Left Channel & Right channel
  AudioEndpointVolume.GetChannelVolumeLevelScalar(0, leftVol);
  AudioEndpointVolume.GetChannelVolumeLevelScalar(1, rightVol);
  refVol := Max(leftVol, rightVol);

  Balance := Min(SessionBalance_Max, tbSessionBalance.Position);
  if Balance < SessionBalance_Mid then // left sided balance
  begin
    leftVol := refVol;
    rightVol := refVol * (Balance / SessionBalance_Mid);
  end
  else if Balance = SessionBalance_Mid then // center balance
  begin
    leftVol := refVol;
    rightVol := refVol;
  end
  else // right sided balance
  begin
    leftVol := refVol * (2.0 - (Balance / SessionBalance_Mid));
    rightVol := refVol;
  end;

  // Set the Volume of Left Channel & Right channel
  AudioEndpointVolume.SetChannelVolumeLevelScalar(0, leftVol, @GUID_TAudioVolume);
  AudioEndpointVolume.SetChannelVolumeLevelScalar(1, rightVol, @GUID_TAudioVolume);

end;

procedure TAudioVolume.tbSessionVolumeChange(Sender: TObject);
begin
  SetSessionVolume((tbSessionVolume.Max - tbSessionVolume.Position) / tbSessionVolume.Max);
end;

procedure TAudioVolume.Timer1Timer(Sender: TObject);
begin
  if not(csDesigning in ComponentState) then
  begin
    if Assigned(pbMasterMeter) then
    begin
      pbMasterMeter.Position := round(GetMasterPeak * 100.0);

      // GetPeakMeter of MasterVolume
      if not GotChannelCount then
      begin
        SetLength(Peaks, 8); // assume max number for channel count
        if GetChannelsMasterPeak(ChannelCount, Peaks) then
        begin
          GotChannelCount := true;
          SetLength(Peaks, ChannelCount);
        end;
      end;

      if GetChannelsMasterPeak(ChannelCount, Peaks) then
      begin
        pbMasterMeterL.Position := round(Peaks[0] * 100.0);
        if ChannelCount > 1 then
          pbMasterMeterR.Position := round(Peaks[1] * 100.0);
      end;
    end;

    // GetPeakMeter of Session
    if Assigned(pbSessionMeter) then
    begin
      pbSessionMeter.Position := round(GetPeak * 100.0);

      // assume max number for channel count
      SetLength(Peaks, 8);
      if GetChannelsPeak(ChannelCount, Peaks) then
        SetLength(Peaks, ChannelCount);

      if GetChannelsPeak(ChannelCount, Peaks) then
      begin
        pbSessionMeterL.Position := round(Peaks[0] * 100.0);
        if ChannelCount > 1 then
          pbSessionMeterR.Position := round(Peaks[1] * 100.0);
      end;
    end;
  end;
end;

procedure TAudioVolume.WMSize(var Message: TWMSize);
begin

  ClientWidth := 116;

  if Height < 177 then
    Height := 177;

  ClientHeight := Height;

  if (not FSessionVolume) then
  begin
    FMGroupBox.Left := 5;
    FMGroupBox.Top := 5;
    FMGroupBox.Width := ClientWidth - 10;
    FMGroupBox.Height := ClientHeight - 10;

    lblMasterCaption.Left := 5;
    lblMasterCaption.Top := 40;
    lblMasterCaption.Width := FMGroupBox.Width + 5;
    lblMasterCaption.Parent := FMGroupBox.Parent;
    lblMasterCaption.Alignment := taCenter;
    lblMasterCaption.AutoSize := true;

    pbMasterMeter.Width := 14;
    pbMasterMeter.Height := IIF(FMasterBalance, FMGroupBox.Height - 142, FMGroupBox.Height - 102);
    pbMasterMeter.Left := 12;
    pbMasterMeter.Top := 82;

    lblMasterPeak.Left := 14;
    lblMasterPeak.Top := 67;

    pbMasterMeterL.Width := 14;
    pbMasterMeterL.Height := IIF(FMasterBalance, FMGroupBox.Height - 142, FMGroupBox.Height - 102);
    pbMasterMeterL.Left := pbMasterMeter.Left + 20;
    pbMasterMeterL.Top := 82;

    lblMasterPeakL.Left := pbMasterMeterL.Left + 4;
    lblMasterPeakL.Top := 67;

    tbMasterVolume.Width := 32;
    tbMasterVolume.Height := pbMasterMeterL.Height + 15;
    tbMasterVolume.Left := pbMasterMeterL.Left + 20;
    tbMasterVolume.Top := 75;

    pbMasterMeterR.Width := 14;
    pbMasterMeterR.Height := IIF(FMasterBalance, FMGroupBox.Height - 142, FMGroupBox.Height - 102);
    pbMasterMeterR.Left := tbMasterVolume.Left + 38;
    pbMasterMeterR.Top := 82;

    lblMasterPeakR.Left := pbMasterMeterR.Left + 3;
    lblMasterPeakR.Top := 67;

    if FMasterBalance then
    begin
      tbMasterBalance.Width := 104;
      tbMasterBalance.Height := 32;
      tbMasterBalance.Left := 6;
      tbMasterBalance.Top := Height - 55;

      lblMasterBalance.Left := tbMasterBalance.Left + 35;
      lblMasterBalance.Top := tbMasterBalance.Top - 12;
    end;

    ckMasterMute.Top := Height - 25;
    ckMasterMute.Left := pbMasterMeter.Left
  end
  else
  begin
    FSGroupBox.Left := 5;
    FSGroupBox.Top := 5;
    FSGroupBox.Width := ClientWidth - 10;
    FSGroupBox.Height := ClientHeight - 10;
    FSGroupBox.Hint := Hint;
    FSGroupBox.ShowHint := true;

    lblSessionCaption.Left := 5;
    lblSessionCaption.Top := 43;
    lblSessionCaption.Width := FSGroupBox.Width + 5;
    lblSessionCaption.AutoSize := true;
    lblSessionCaption.Parent := FSGroupBox.Parent;
    lblSessionCaption.Alignment := taCenter;

    Image := TImage.Create(FSGroupBox);
    Image.AutoSize := true;
    Image.Parent := FSGroupBox;
    Image.Picture.Icon := Icon;
    if Assigned(Icon) then
      Image.Left := (ClientWidth - Icon.Width) div 2;
    Image.Top := 5;
    Image.Show;

    pbSessionMeter.Width := 14;
    pbSessionMeter.Height := IIF(FSessionBalance, FSGroupBox.Height - 142, FSGroupBox.Height - 102);
    pbSessionMeter.Left := 12;
    pbSessionMeter.Top := 82;

    lblSessionPeak.Left := 14;
    lblSessionPeak.Top := 67;

    pbSessionMeterL.Width := 14;
    pbSessionMeterL.Height := IIF(FSessionBalance, FSGroupBox.Height - 142,
      FSGroupBox.Height - 102);
    pbSessionMeterL.Left := pbSessionMeter.Left + 20;
    pbSessionMeterL.Top := 82;

    lblSessionPeakL.Left := pbSessionMeterL.Left + 4;
    lblSessionPeakL.Top := 67;

    tbSessionVolume.Width := 32;
    tbSessionVolume.Height := pbSessionMeterL.Height + 15;
    tbSessionVolume.Left := pbSessionMeterL.Left + 20;
    tbSessionVolume.Top := 75;

    pbSessionMeterR.Width := 14;
    pbSessionMeterR.Height := IIF(FSessionBalance, FSGroupBox.Height - 142,
      FSGroupBox.Height - 102);
    pbSessionMeterR.Left := tbSessionVolume.Left + 38;
    pbSessionMeterR.Top := 82;

    lblSessionPeakR.Left := pbSessionMeterR.Left + 3;
    lblSessionPeakR.Top := 67;

    if FSessionBalance then
    begin
      if not Assigned(tbSessionBalance) then
        SetSessionBalance(FSessionBalance);
      tbSessionBalance.Width := 104;
      tbSessionBalance.Height := 32;
      tbSessionBalance.Left := 6;
      tbSessionBalance.Top := Height - 55;

      lblSessionBalance.Left := tbSessionBalance.Left + 35;
      lblSessionBalance.Top := tbSessionBalance.Top - 12;
    end;

    ckSessionMute.Top := Height - 25;
    ckSessionMute.Left := pbSessionMeter.Left
  end;
end;

function TAudioVolume.GetMasterBalance: DWORD;
var
  HR: HResult;
  Channels: uint;
  Volume: TVolume;
  leftVol, rightVol: single;

begin
  Result := MasterBalance_Mid;

  // ** followings are added sentences for Audio Mixer API
  if Win32MajorVersion < 6 then // for Windows XP
  begin
    if not FMixerAPIReady then
      exit;

    if (FChannels < 2) or ((MasterVol.fdwControl and MIXERCONTROL_CONTROLF_UNIFORM)
        = MIXERCONTROL_CONTROLF_UNIFORM) then
      exit;
    if GetMixerVolume(FMixerObj, FChannels, MasterVol, Volume) = MMSYSERR_NOERROR then
    begin
      leftVol := Volume[0] / MasterVol.Bounds.dwMaximum;
      rightVol := Volume[1] / MasterVol.Bounds.dwMaximum;
      // Get the Balance value according to the volume of left & right channel
      if leftVol > rightVol then
        Result := round((rightVol / leftVol) * MasterBalance_Mid)
      else if leftVol < rightVol then
        Result := round((2.0 - (leftVol / rightVol)) * MasterBalance_Mid);
    end;

    exit;
  end; // ** end of added sentences

  if AudioEndpointVolume = nil then
    exit;

  HR := AudioEndpointVolume.GetChannelCount(Channels);
  if HR <> S_OK then
    exit
  else if Channels < 2 then
    exit
  else
  begin
    // Get the current Volume of Left Channel & Right channel
    AudioEndpointVolume.GetChannelVolumeLevelScalar(0, leftVol);
    AudioEndpointVolume.GetChannelVolumeLevelScalar(1, rightVol);

    // Get the Balance value according to the volume of left & right channel
    if leftVol > rightVol then
      Result := round((rightVol / leftVol) * MasterBalance_Mid)
    else if leftVol < rightVol then
      Result := round((2.0 - (leftVol / rightVol)) * MasterBalance_Mid);
  end;
end;

function TAudioVolume.GetMasterPeak: single;
var
  HR: HResult;
  fPeak: single;
begin
  Result := 0;

  if MasterMeter = nil then
    exit;

  try
    HR := MasterMeter.GetPeakValue(fPeak);
    if HR = S_OK then
      Result := fPeak;
  except

  end;

end;

function TAudioVolume.GetPeak: single;
var
  HR: HResult;
  fPeak: single;
begin
  Result := 0;

  if PeakMeter = nil then
    exit;

  try
    HR := PeakMeter.GetPeakValue(fPeak);
    if Succeeded(HR) then
      Result := fPeak;
  except

  end;

end;

function TAudioVolume.GetProcessIdFromModuleName(ModuleName: PWideChar): LongWord;
var
  ToolSnap: LongWord;
  ProcessEntry: ProcessEntry32W;
  ModuleName16: WideString;
  bCont: LongBool;
begin
  Result := 0;
  ModuleName16 := ModuleName;
  ToolSnap := CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  ProcessEntry.dwSize := SizeOf(ProcessEntry);
  bCont := Process32FirstW(ToolSnap, ProcessEntry);
  while bCont do
  begin
    if ProcessEntry.szExeFile = ModuleName16 then
      Result := ProcessEntry.th32ProcessID;
    ProcessEntry.dwSize := SizeOf(ProcessEntry);
    bCont := (Result = 0) and Process32NextW(ToolSnap, ProcessEntry);
  end;
  CloseHandle(ToolSnap);

end;

function TAudioVolume.GetProcessPathByHandle(hProcess: THandle): WideString;
var
  QueryFullProcessImageName: function(hProcess: THandle; dwFlags: DWORD; ImageFileName: PWideChar;
    dwSize: PDWord): Bool; stdcall;

  FileName :array [0 .. MAX_PATH - 1] of WideChar;
  BufferSize : DWORD;
begin
  Result := '';
  // Windows Vista or higher
  @QueryFullProcessImageName := GetProcAddress(LoadLibrary('kernel32.dll'),
    'QueryFullProcessImageNameW');
  if Assigned(QueryFullProcessImageName) then
  begin
    BufferSize := MAX_PATH;
    if (QueryFullProcessImageName(hProcess, 0, @FileName[0], @BufferSize)) then
    begin
      Result := PWideChar(@FileName[0]);
      exit;
    end;
  end;
end;

function TAudioVolume.GetProcessPathByPID(PID: DWORD): WideString;
var
  hProcess: THandle;
begin

  Result := '';
  hProcess := OpenProcess(PROCESS_QUERY_INFORMATION or PROCESS_VM_READ, false, PID);
  if (hProcess <> 0) and (hProcess <> INVALID_HANDLE_VALUE) then
    try
      Result := GetProcessPathByHandle(hProcess);
    finally
      CloseHandle(hProcess);
    end;
end;

function TAudioVolume.GetSessionCount(const AudioDevice: IMMDevice): integer;
var
  SessionCount: integer;
  HR: HResult;
begin

  Result := 0;

  HR := AudioDevice.Activate(IID_IAudioSessionManager, CLSCTX_INPROC_SERVER, nil, IUnknown
      (SessionManager));
  if Succeeded(HR) then
  begin
    HR := SessionEnumerator.GetCount(SessionCount);
    if Succeeded(HR) then
      Result := SessionCount;
  end;
end;

function TAudioVolume.GetSessionPID(Index: integer; const AudioDevice: IMMDevice): DWORD;
var
  SessionControl2: IAudioSessionControl2;
  State, UseState: uint;
  PID: DWORD;
  HR: HResult;
begin
  Result := 0;

  HR := AudioDevice.Activate(IID_IAudioSessionManager, CLSCTX_INPROC_SERVER, nil, IUnknown
      (SessionManager));
  if Succeeded(HR) then
  begin
    HR := SessionEnumerator.GetSession(Index, SessionControl);
    if Succeeded(HR) then
    begin
      HR := SessionControl.QueryInterface(IAudioSessionControl2, SessionControl2);
      if Succeeded(HR) then
      begin
        HR := SessionControl2.GetProcessID(PID);
        if Succeeded(HR) then
        begin
          HR := SessionControl2.GetState(State);
          if Succeeded(HR) then
          begin
            if UseSessionInactive then
              UseState := DWORD(AudioSessionStateExpired)
            else
              UseState := DWORD(AudioSessionStateActive);

            if UseState = DWORD(AudioSessionStateExpired) then
            begin
              if State <> UseState then
                Result := PID
            end
            else if State = UseState then
              Result := PID
          end;
        end;
      end;
      SessionControl2 := nil;
    end;
  end;
end;

procedure TAudioVolume.SetSessionBalance(const Value: boolean);
begin

  FSessionBalance := Value;
  if FSessionBalance and Assigned(FSGroupBox) then
  begin
    tbSessionBalance := TTrackBar.Create(self);
    tbSessionBalance.Parent := FSGroupBox.Parent;
    tbSessionBalance.Orientation := trHorizontal;
    tbSessionBalance.Max := SessionBalance_Max;
    tbSessionBalance.Min := SessionBalance_Min;
    tbSessionBalance.TickStyle := tsAuto;
    tbSessionBalance.Frequency := 3;
    tbSessionBalance.Position := SessionBalance_Mid;

    tbSessionBalance.OnChange := tbSessionBalanceChange;

    lblSessionBalance := TLabel.Create(self);
    lblSessionBalance.Parent := FSGroupBox.Parent;
    lblSessionBalance.Caption := 'Balance';
  end
  else
  begin
    if Assigned(lblSessionBalance) then
    begin
      FreeAndNil(tbSessionBalance);
      FreeAndNil(lblSessionBalance);
    end;
  end;
  PostMessage(Handle, WM_SIZE, 0, 0);
end;

procedure TAudioVolume.SetSessionVolume(Value: single);
begin
  if Value < 0 then
    exit;
  if Value > 1.0 then
    exit;
  if Win32MajorVersion < 6 then
    exit;
  if SimpleAudioVolume = nil then
    exit;

  MySessionVolSet := true;
  MyVolVal := Value;
  SimpleAudioVolume.SetMasterVolume(Value, @GUID_TAudioVolume);
end;

procedure TAudioVolume.SetSessionComponents(Value: boolean);
begin

  if Value = FSessionVolume then
    exit;

  FSessionVolume := Value;

  if FSessionVolume then
  begin
    FSGroupBox := TGroupBox.Create(self);
    FSGroupBox.Parent := self;

    FreeAndNil(pbMasterMeter);
    FreeAndNil(lblMasterPeak);
    FreeAndNil(pbMasterMeterL);
    FreeAndNil(lblMasterPeakL);
    FreeAndNil(tbMasterBalance);
    FreeAndNil(tbMasterVolume);
    FreeAndNil(ckMasterMute);
    FreeAndNil(pbMasterMeterR);
    FreeAndNil(lblMasterPeakR);
    FreeAndNil(lblMasterCaption);
    FreeAndNil(FMGroupBox);

    lblSessionCaption := TLabel.Create(self);
    lblSessionCaption.Parent := FSGroupBox.Parent;
    lblSessionCaption.AutoSize := true;
    lblSessionCaption.Alignment := taCenter;
    lblSessionCaption.Enabled := true;

    pbSessionMeter := TProgressBar.Create(self);
    pbSessionMeter.Parent := FSGroupBox.Parent;
    pbSessionMeter.Orientation := pbVertical;
    pbSessionMeter.Max := 100;
    pbSessionMeter.Min := 0;
    pbSessionMeter.Smooth := true;

    lblSessionPeak := TLabel.Create(self);
    lblSessionPeak.Parent := FSGroupBox.Parent;
    lblSessionPeak.Caption := 'M';

    pbSessionMeterL := TProgressBar.Create(self);
    pbSessionMeterL.Parent := FSGroupBox.Parent;
    pbSessionMeterL.Orientation := pbVertical;
    pbSessionMeterL.Max := 100;
    pbSessionMeterL.Min := 0;
    pbSessionMeterL.Smooth := true;

    lblSessionPeakL := TLabel.Create(self);
    lblSessionPeakL.Parent := FSGroupBox.Parent;
    lblSessionPeakL.Caption := 'L';

    tbSessionVolume := TTrackBar.Create(self);
    tbSessionVolume.Parent := FSGroupBox.Parent;
    tbSessionVolume.Orientation := trVertical;
    tbSessionVolume.Max := 32;
    tbSessionVolume.Min := 0;
    tbSessionVolume.TickStyle := tsAuto;
    tbSessionVolume.Frequency := 3;
    tbSessionVolume.Position := 0;

    ckSessionMute := TCheckbox.Create(self);
    ckSessionMute.Parent := FSGroupBox.Parent;
    ckSessionMute.Caption := 'Mute';

    pbSessionMeterR := TProgressBar.Create(self);
    pbSessionMeterR.Parent := FSGroupBox.Parent;
    pbSessionMeterR.Orientation := pbVertical;
    pbSessionMeterR.Max := 100;
    pbSessionMeterR.Min := 0;
    pbSessionMeterR.Smooth := true;

    lblSessionPeakR := TLabel.Create(self);
    lblSessionPeakR.Parent := FSGroupBox.Parent;
    lblSessionPeakR.Caption := 'R';

    tbSessionVolume.Position := round((1.0 - GetVolume) * tbSessionVolume.Max);
    tbSessionVolume.OnChange := tbSessionVolumeChange;
    ckSessionMute.Checked := IsMuted;
    ckSessionMute.OnClick := ckSessionMuteClick;
    PostMessage(Handle, WM_SIZE, 0, 0);
  end
  else
  begin
    FreeAndNil(pbSessionMeter);
    FreeAndNil(lblSessionPeak);
    FreeAndNil(pbSessionMeterL);
    FreeAndNil(lblSessionPeakL);
    FreeAndNil(tbSessionBalance);
    FreeAndNil(tbSessionVolume);
    FreeAndNil(ckSessionMute);
    FreeAndNil(pbSessionMeterR);
    FreeAndNil(lblSessionPeakR);
    FreeAndNil(lblSessionCaption);
    FreeAndNil(Image);
    FreeAndNil(FSGroupBox);
  end;

end;

procedure TAudioVolume.SetSessionInactive(const Value: boolean);
begin

  FSessionInactive := Value;
end;

procedure TAudioVolume.SetSessionMute(Value: boolean);
begin
  if Win32MajorVersion < 6 then
    exit;
  if SimpleAudioVolume = nil then
    exit;

  MyMuteSet := true;
  MyMuteVal := Value;
  SimpleAudioVolume.SetMute(Value, @GUID_TAudioVolume);
  ckSessionMute.Checked := Value;
end;

function TAudioVolume.SetSimpleAudioVolumeFromHandle(const AudioDevice: IMMDevice; HWND: uint;
  Index: integer; out SimpleAudio: ISimpleAudioVolume; out currguid: TGUID): boolean;
var
  PID: uint;
begin

  GetWindowThreadProcessId(HWND, PID);
  Result := SetSimpleAudioVolumeFromPid(AudioDevice, PID, Index, SimpleAudio, currguid);
end;

function TAudioVolume.SetSimpleAudioVolumeFromModuleName(const AudioDevice: IMMDevice;
  ModuleName: PWideChar; Index: integer; out SimpleAudio: ISimpleAudioVolume; out currguid: TGUID)
  : boolean;
var
  PID: LongWord;
begin

  PID := GetProcessIdFromModuleName(ModuleName);
  Result := SetSimpleAudioVolumeFromPid(AudioDevice, PID, Index, SimpleAudio, currguid);
end;

function TAudioVolume.SetSimpleAudioVolumeFromPid(const AudioDevice: IMMDevice; ProcId: uint;
  Index: integer; out SessionAudioVolume: ISimpleAudioVolume; out currguid: TGUID): boolean;
var
  SessionControl2: IAudioSessionControl2;
  pEnumerator: IAudioSessionEnumerator;
  HR: HResult;
  PID: uint;
  State, UseState: uint;

begin
  // get AudioSessionManager
  HR := AudioDevice.Activate(IID_IAudioSessionManager, CLSCTX_INPROC_SERVER, nil, IUnknown
      (SessionManager));
  if Succeeded(HR) then
  begin
    HR := SessionManager2.GetSessionEnumerator(pEnumerator);
    if Succeeded(HR) then
    begin
      HR := pEnumerator.GetSession(Index, SessionControl);
      if Succeeded(HR) then
      begin
        HR := SessionControl.QueryInterface(IAudioSessionControl2, SessionControl2);
        if Succeeded(HR) then
        begin
          HR := SessionControl2.IsSystemSoundsSession;
          if Succeeded(HR) then
          begin
            FAudioSessionList.IsSystemSoundsSession := HR;
            FAudioSessionList.ActiveSession := intToStr(Index);
          end;

          HR := SessionControl2.GetProcessID(PID);
          if Succeeded(HR) then
          begin
            FAudioSessionList.ProcessID := intToStr(PID);
            HR := SessionControl2.GetState(State);
            if Succeeded(HR) then
            begin
              if PID = ProcId then
              begin
                if UseSessionInactive then
                  UseState := DWORD(AudioSessionStateExpired)
                else
                  UseState := DWORD(AudioSessionStateActive);

                if UseState = DWORD(AudioSessionStateExpired) then
                begin
                  if State <> UseState then
                    SetDiplayNameAndIcon(PID, Index, SessionAudioVolume, SessionControl,
                      SessionControl2, State);
                end
                else
                begin
                  if State = UseState then
                    SetDiplayNameAndIcon(PID, Index, SessionAudioVolume, SessionControl,
                      SessionControl2, State);
                end;
              end;
            end;
          end;
        end;
        SessionControl2 := nil;
      end;
    end;
  end;

  Result := Assigned(SessionAudioVolume);
end;

function TAudioVolume.GetAudioDevice: TStrings;
var
  N: integer;
  FDeviceList: array of TDeviceInfo;
  Items: TStringList;

begin
  Items := TStringList.Create;
  FAudioDevice := TStringList.Create;
  if FDeviceCount > 0 then
  begin
    SetLength(FDeviceList, FDeviceCount);
    for N := 0 to (FDeviceCount - 1) do
    begin
      FDeviceList[N] := GetDeviceInfo2(DeviceCollection, N);
{$IFDEF UNICODE}
      ExtractStrings(['('], [], PWideChar(WideString(FDeviceList[N].DeviceName)), Items);
{$ELSE}
      ExtractStrings(['('], [], PChar(AnsiString(FDeviceList[N].DeviceName)), Items);
{$ENDIF}
      FAudioDevice.Add(Items[0]);
      Items.Clear;
    end;
    Items.free;
  end;

  Result := FAudioDevice;

end;

function TAudioVolume.GetChannelsMasterPeak(var Channels: uint; var Peaks: array of single)
  : boolean;
var
  HR: HResult;
begin
  Result := false;
  Channels := 0;

  if MasterMeter = nil then
    exit;

  try
    // ** The vaule of Channels, returned by GetMeteringChannelCount
    // for non-active(=not playing state) audio session, is 0
    HR := MasterMeter.GetMeteringChannelCount(Channels);
    if Succeeded(HR) and (Channels > 0) then
    begin
      HR := MasterMeter.GetChannelsPeakValues(Channels, @Peaks[0]);

      if Succeeded(HR) then
        Result := true;
    end;

  except

  end;
end;

function TAudioVolume.GetChannelsPeak(var Channels: uint; var Peaks: array of single): boolean;
var
  HR: HResult;
begin
  Result := false;
  Channels := 0;

  if PeakMeter = nil then
    exit;

  try
    // ** The vaule of Channels, returned by GetMeteringChannelCount
    // for non-active(=not playing state) audio session, is 0
    HR := PeakMeter.GetMeteringChannelCount(Channels);
    if Succeeded(HR) and (Channels > 0) then
    begin
      HR := PeakMeter.GetChannelsPeakValues(Channels, @Peaks[0]);

      if Succeeded(HR) then
        Result := true;
    end else
    begin
      if Assigned(pbSessionMeter) then
      begin
        pbSessionMeterL.Position := round(Peaks[0]);
        pbSessionMeterR.Position := round(Peaks[1]);
      end;
    end;

  except

  end;
end;

function TAudioVolume.GetDevice: IMMDevice;
begin

  Result := FDevice;
end;

function TAudioVolume.GetIcon: TIcon;
begin

  Result := FIcon;
end;

end.
