{ Demo program to test TAudioVolume


  Author : Silhwan Hyun  (hyunsh@hanafos.com)
  Contributor : Emil Weiss (weiss.em@arcor.de)
  Contributor : HolgerX (https://www.delphipraxis.net/members/19009-holgerx.html)

  Revision History
  -----------------------------
  v0.9.8 (6 Feb 2017)
  - Adapted to TAudioVolume v0.9.8

  v0.9.7
  - Adapted to TAudioVolume v0.9.7

  v0.9.6    (Pending)

  v0.9.5    (06 Jun 2014)
  - Adapted to TAudioVolume v0.9.5
  ( Modification to display application's audio output level );

  v0.9.4    (13 Mar 2012)
  - Adapted to TAudioVolume v0.9.4

  V0.9.3    (15 Aug 2011)
  - Adapted to TAudioVolume v0.9.3
  - Added peak meters for Output device
  - Added function to play MIDI files to show the operation of peak meters

  V0.9.2    (12 Aug 2011)
  - Adapted to TAudioVolume v0.9.2

  V0.9.1    (03 Aug 2011)
  - Added a edit box to show you the Operation Mode.
  - Replaced custom TrackBars(=> TMyTrackBar) with standard TrackBars.

  V0.9.0    (31 Jul 2011)
  - Initial Release
}

unit VolumeTest_Eng;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms, ActiveX,
  Dialogs, StdCtrls, ComCtrls, AudioVolume, ExtCtrls, MPlayer, SimpleAudioVolume, MMDevice;

type
  TForm1 = class(TForm)
    btnClose: TButton;
    cbOutputDevices: TComboBox;
    Label1: TLabel;
    Label4: TLabel;
    edMode: TEdit;
    MediaPlayer1: TMediaPlayer;
    btnOpen: TButton;
    OpenDialog1: TOpenDialog;
    lblBPS: TLabel;
    lblSPS: TLabel;
    edOutputAttrib: TEdit;
    Label7: TLabel;
    Label6: TLabel;
    Label8: TLabel;
    Label5: TLabel;
    Label9: TLabel;
    Label10: TLabel;
    CheckBox1: TCheckBox;
    btnActiveSession: TButton;
    gbSession: TGroupBox;
    meSession: TMemo;
    AudioVolume1: TAudioVolume;
    procedure FormShow(Sender: TObject);
    procedure btnCloseClick(Sender: TObject);
    procedure btnOpenClick(Sender: TObject);
    procedure cbOutputDevicesChange(Sender: TObject);
    procedure AudioDefaultDeviceChange(NewDefaultDevice: TDeviceInfo);
    procedure AudioDeviceStateChange(DeviceId: string; NewDeviceState: Cardinal);
    procedure SessionStateEvent(Sender : TAudioVolume; NewState: integer);
    procedure SessionDisconnectedEvent(Sender : TAudioVolume; DisconnectReason: uint);
    procedure SessionVolumeChanged(ByMyAction: Boolean; Volume: Single;
      Mute: Boolean);
    procedure CheckBox1Click(Sender: TObject);
    procedure btnActiveSessionClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure AudioVolume1MasterVolumeChanged(ByMyAction: Boolean; Volume: Single; Mute: Boolean);
  private
    { Private declarations }
    SimpleVolume: ISimpleAudioVolume;
    SessionGUID: TGUID;
    OldProcId: array of DWORD;
    MaxWidth: Integer;
    procedure Display(const S: string);
    procedure ShowOutputAttribute;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;
  AudioVolumeArr: array of TAudioVolume;

implementation

{$R *.dfm}

var
  WasPlaying: boolean = false;

procedure TForm1.AudioDefaultDeviceChange(NewDefaultDevice: TDeviceInfo);
var
  S: string;
  i: integer;
begin
  S := 'Default audio endpoint device for output was changed to,' + chr(10)
    + ' Name : ' + NewDefaultDevice.DeviceName + chr(10) + '  Id : ' + NewDefaultDevice.DeviceId;
  Application.MessageBox(PChar(S), 'Information', MB_OK);

  for i := 0 to cbOutputDevices.Items.Count - 1 do
    if cbOutputDevices.Items[i] = NewDefaultDevice.DeviceName then
    begin
      cbOutputDevices.ItemIndex := i;
      break;
    end;

end;

procedure TForm1.AudioDeviceStateChange(DeviceId: string; NewDeviceState: Cardinal);
var
  S: string;
  S2: string;
begin
  case NewDeviceState of
    DEVICE_STATE_ACTIVE:
      S2 := '"Active"';
    DEVICE_STATE_DISABLED:
      S2 := '"Disabled"';
    DEVICE_STATE_NOTPRESENT:
      S2 := '"Not present"';
    DEVICE_STATE_UNPLUGGED:
      S2 := '"Unplugged"';
    DEVICE_STATE_ADDED:
      S2 := '"Added"';
    DEVICE_STATE_REMOVED:
      S2 := '"Removed"';
  end;
  S := 'Following Device state was changed to ' + S2 + ' state,' + chr(10) + '  Id : ' + DeviceId;
end;

procedure TForm1.AudioVolume1MasterVolumeChanged(ByMyAction: Boolean; Volume: Single;
  Mute: Boolean);
var
  i: Integer;
begin

  for i := 0 to Length(AudioVolumeArr) -1 do
  begin
    AudioVolumeArr[i].SetSessionMute(Mute);
    AudioVolumeArr[i].SetSessionVolume(Volume);
  end;
end;

procedure TForm1.SessionVolumeChanged(ByMyAction: Boolean; Volume: Single;
  Mute: Boolean);
begin
  //
end;

procedure TForm1.btnActiveSessionClick(Sender: TObject);
begin
  if btnActiveSession.Caption = '>' then
  begin
    if MaxWidth < 680 then
      Width := 680
    else
    Width := MaxWidth;

    btnActiveSession.Caption := '<'
  end
  else
  begin
    Width := 370;
    btnActiveSession.Caption := '>'
  end;
end;

procedure TForm1.btnCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TForm1.ShowOutputAttribute;
begin
  edOutputAttrib.Text := 'Channels: ' + intToStr(AudioVolume1.MasterChannels);
  if not AudioVolume1.WindowXPMode then
    edOutputAttrib.Text := edOutputAttrib.Text + ',  ' + intToStr
      (AudioVolume1.DeviceMixFormat.nSamplesPerSec) + ' SPS,  ' + intToStr
      (AudioVolume1.DeviceMixFormat.wBitsPerSample) + ' BPS';
end;

procedure TForm1.cbOutputDevicesChange(Sender: TObject);
begin
  AudioVolume1.SelectAudioDevice(cbOutputDevices.ItemIndex);
  ShowOutputAttribute;

end;

procedure TForm1.CheckBox1Click(Sender: TObject);
begin

  AudioVolume1.MasterBalance := not AudioVolume1.MasterBalance;
end;

procedure TForm1.Display(const S: string);
begin

  meSession.Lines.Add(S);
end;

procedure TForm1.FormResize(Sender: TObject);
begin

  gbSession.Width := (Width - gbSession.left) - 25;
end;

procedure TForm1.FormShow(Sender: TObject);
var
  N, i, j: Cardinal;
  Count: integer;
  ProcId: DWord;
  Offset, space: Integer;

  tmpAudioVolume : TAudioVolume;
begin
  Offset := 0;

  if not AudioVolume1.Ready then
  begin
    MessageBox(Self.Handle, 'Failed at initialization.', 'Confirm', MB_OK);
    exit;
  end;

  N := AudioVolume1.AudioDeviceCount;
  if N = 0 then
  begin
    MessageBox(Self.Handle, 'No Device to control.', 'Confirm', MB_OK);
    exit;
  end;

  i := 0;
  for N := 0 to AudioVolume1.AudioDeviceCount - 1 do
    if AudioVolume1.AudioDeviceInfo(N).DeviceId = AudioVolume1.DefaultAudioDevice.DeviceId then
    begin
      i := N; // the index to select default audio device
      break;
    end;

  if AudioVolume1.WindowXPMode then
  begin
    AudioVolume1.SelectAudioDevice(i);

    edMode.Text := 'Audio Mixer API';
    lblSPS.Caption := '* Shows only Channels at Windows XP.';
    lblBPS.Visible := false;
  end
  else
  begin
    AudioVolume1.SelectAudioDevice(i);
    edMode.Text := 'Core Audio API';
  end;

  ShowOutputAttribute;

  for N := 0 to AudioVolume1.AudioDeviceCount - 1 do
    cbOutputDevices.Items.Add(AudioVolume1.AudioDeviceInfo(N).DeviceName);
  cbOutputDevices.ItemIndex := i;

  MediaPlayer1.Width := 130;
  MediaPlayer1.Height := 25;

  Count := AudioVolume1.GetSessionCount(AudioVolume1.Device);
  SetLength(OldProcId, Count + 1);

  space := 6;

  for i := 0 to (Count - 1) do
  begin
    tmpAudioVolume := TAudioVolume.Create(self);
    if i = 0 then
    begin
      tmpAudioVolume.Parent := self;
      tmpAudioVolume.Left := 177;
      tmpAudioVolume.Top := 132;
      tmpAudioVolume.Height := 211;
      tmpAudioVolume.Visible := true;
    end else
    tmpAudioVolume.Parent := gbSession;

    tmpAudioVolume.SessionVolume := true;
    tmpAudioVolume.SessionBalance := false;
    tmpAudioVolume.OnSessionStateEvent := SessionStateEvent;
    tmpAudioVolume.OnSessionDisconnected := SessionDisconnectedEvent;
    tmpAudioVolume.OnSessionVolumeChanged := SessionVolumeChanged;
    tmpAudioVolume.UseSessionInactive := true;
    if i = 0 then
      ProcId := 0
    else
    ProcId := tmpAudioVolume.GetSessionPID(i, AudioVolume1.Device);

    tmpAudioVolume.SetupPeakMeter(ProcId);

    SimpleVolume := nil;

    if tmpAudioVolume.SetSimpleAudioVolumeFromPid(AudioVolume1.Device, ProcId, i,
      SimpleVolume, SessionGUID) then
    begin
      //meSession.Clear;
      Display('Active Session = ' + tmpAudioVolume.AudioSessionInfo(i).ActiveSession);
      Display('DisplayName = ' + tmpAudioVolume.AudioSessionInfo(i).DisplayName);
      Display('StateSession = ' + tmpAudioVolume.AudioSessionInfo(i).StateSession + #13#10);
      Display('IconPath = ' + tmpAudioVolume.AudioSessionInfo(i).IconPath + #13#10);
      Display('SessionIdentifier = ' + tmpAudioVolume.AudioSessionInfo(i).SessionIdentifier + #13#10);
      Display('ProcessID = ' + tmpAudioVolume.AudioSessionInfo(i).ProcessID);
      Display('IsSystemSoundsSession = ' + intToStr(tmpAudioVolume.AudioSessionInfo(i).IsSystemSoundsSession) + #13#10);
      Display('SessionGuid = ' + GUIDToString(tmpAudioVolume.AudioSessionInfo(i).SessionGUID));
      if tmpAudioVolume.AudioSessionInfo(i).DisplayName <> '' then
        Display('******************************' + #13#10)
      else
      Display('*******SESSION REMOVED********' + #13#10)
    end;

    if Assigned(SimpleVolume) then
    begin
      for j := 0 to length(OldProcId) - 1 do
      begin
        // not remove SystemSounds ProcId is NULL
        if ((ProcId = OldProcId[j]) and not (ProcId = 0)) or
          (tmpAudioVolume.AudioSessionInfo(i).DisplayName = '') then
        begin
          FreeAndNil(tmpAudioVolume);
          break;
        end;
      end;
      if i > 0 then
      begin
        if Assigned(tmpAudioVolume) then
        begin
          tmpAudioVolume.Left := 6 + Offset;
          tmpAudioVolume.Top := 280;
          tmpAudioVolume.Height := 211;

          tmpAudioVolume.Visible := true;

          inc(Offset, 116);
          MaxWidth := (362 + (Offset + space));
          inc(space, 6);
          OldProcId[i] := ProcId;
        end;
      end;
    end else
      FreeAndNil(tmpAudioVolume);

    if Assigned(tmpAudioVolume) then
    begin
      SetLength(AudioVolumeArr, length(AudioVolumeArr)+1);
      AudioVolumeArr[length(AudioVolumeArr)-1] := tmpAudioVolume;
    end;
  end;

//{$IFDEF VER210} // Delphi 2010 ?
//  ReportMemoryLeaksOnShutdown := DebugHook <> 0;
//{$ENDIF}
end;

procedure TForm1.btnOpenClick(Sender: TObject);
begin
  if OpenDialog1.Execute then
  begin
    MediaPlayer1.FileName := OpenDialog1.FileName;
    MediaPlayer1.Open;
  end;
end;

// Notifies the client that the session has been disconnected.
procedure TForm1.SessionDisconnectedEvent(Sender: TAudioVolume; DisconnectReason: uint);
var
  i : integer;
begin
  for i := 0 to Length(AudioVolumeArr) -1 do
  begin
    if AudioVolumeArr[i] = Sender then
    begin
      ShowMessage(AudioVolumeArr[i].AudioSessionInfo(i).DisplayName + #13#10 +
        'Index ' + IntToStr(i) + ' Audio session disconnected ' + #13#10 +
        AudioVolumeArr[i].AudioSessionInfo(i).SessionDisconnected);
      break;
    end;
  end;

end;

// Notifies the client that the stream-activity state of the session has changed.
procedure TForm1.SessionStateEvent(Sender: TAudioVolume;
  NewState: integer);
var
  i : integer;
begin
  for i := 0 to Length(AudioVolumeArr) -1 do
  begin
    if AudioVolumeArr[i] = Sender then
    begin
//      ShowMessage(AudioVolumeArr[i].AudioSessionInfo(i).DisplayName + #13#10 +
//      'Index ' + IntToStr(i) + ' hat jetzt State ' + IntToStr(NewState));
      break;
    end;
  end;
end;

end.
