unit MMNotificationClient;

interface

uses
  Windows, Messages, Classes, PropSys2, Constants;

const
  IID_IMMNotificationClient: TGUID = '{7991EEC9-7E89-4D85-8390-6C703CEC60C0}';

type
  IMMNotificationClient = interface(IUnknown)
    ['{7991EEC9-7E89-4D85-8390-6C703CEC60C0}']
    function OnDeviceStateChanged(DeviceId: LPCWSTR; NewState: DWORD): HRESULT; stdcall;
    function OnDeviceAdded(DeviceId: LPCWSTR): HRESULT; stdcall;
    function OnDeviceRemoved(DeviceId: LPCWSTR): HRESULT; stdcall;
    function OnDefaultDeviceChanged(Flow: EDataFlow; Role: ERole; NewDefaultDeviceId: LPCWSTR)
      : HRESULT; stdcall;
    function OnPropertyValueChanged(DeviceId: LPCWSTR; const Key: PROPERTYKEY): HRESULT; stdcall;
  end;

  TMMNotificationClient = class(TInterfacedObject, IMMNotificationClient)
  private
    MsgHandle: HWND;
    tmpDefaultId: WideString;
  public
    function OnDeviceStateChanged(DeviceId: LPCWSTR; NewState: DWORD): HRESULT; stdcall;
    function OnDeviceAdded(DeviceId: LPCWSTR): HRESULT; stdcall;
    function OnDeviceRemoved(DeviceId: LPCWSTR): HRESULT; stdcall;
    function OnDefaultDeviceChanged(Flow: EDataFlow; Role: ERole; NewDefaultDeviceId: LPCWSTR)
      : HRESULT; stdcall;
    function OnPropertyValueChanged(DeviceId: LPCWSTR; const Key: PROPERTYKEY): HRESULT; stdcall;
    procedure SetMsgHandle(WinHandle: HWND);
  end;

implementation

function TMMNotificationClient.OnDeviceStateChanged(DeviceId: LPCWSTR; NewState: DWORD): HRESULT;
  stdcall;
begin
  PostMessage(MsgHandle, WM_DeviceStateChange, WPARAM(DeviceId), LPARAM(NewState));

  Result := S_OK;
end;

function TMMNotificationClient.OnDeviceAdded(DeviceId: LPCWSTR): HRESULT; stdcall;
begin
  PostMessage(MsgHandle, WM_DeviceAdded, WPARAM(DeviceId), 0);

  Result := S_OK;
end;

function TMMNotificationClient.OnDeviceRemoved(DeviceId: LPCWSTR): HRESULT; stdcall;
begin
  PostMessage(MsgHandle, WM_DeviceRemoved, WPARAM(DeviceId), 0);

  Result := S_OK;
end;

function TMMNotificationClient.OnDefaultDeviceChanged(Flow: EDataFlow; Role: ERole;
  NewDefaultDeviceId: LPCWSTR): HRESULT; stdcall;
var
  pw: pWideChar;
begin
  if (Flow = eRender) and (Role = eMultimedia) then
  begin
    // the content pointed by NewDefaultDeviceId is unavailable after exiting this function
    pw := NewDefaultDeviceId;
    tmpDefaultId := WideString(pw);
    pw := pWideChar(tmpDefaultId);
    PostMessage(MsgHandle, WM_DefaultDeviceChange, WPARAM(pw), 0);
  end;

  Result := S_OK;
end;

function TMMNotificationClient.OnPropertyValueChanged(DeviceId: LPCWSTR; const Key: PROPERTYKEY)
  : HRESULT; stdcall;
begin
  Result := S_OK;
end;

procedure TMMNotificationClient.SetMsgHandle(WinHandle: HWND);
begin
  MsgHandle := WinHandle;
end;

end.
