unit AudioSessionEvents;

interface

uses
  Windows, Messages, SysUtils, Classes, Constants;

type
  TAudioSessionState = (AudioSessionStateInactive, AudioSessionStateActive,
    AudioSessionStateExpired);

  TAudioSessionDisconnected = (DisconnectReasonDeviceRemoval, DisconnectReasonServerShutdown,
    DisconnectReasonFormatChanged, DisconnectReasonSessionLogoff,
    DisconnectReasonSessionDisconnected, DisconnectReasonExclusiveModeOverride);
const
  IID_IAudioSessionEvents: TGUID = '{24918ACC-64B3-37C1-8CA9-74A66E9957A8}';

  cAudioSessionState: Array [TAudioSessionState] of String = ('AudioSessionStateInactive',
    'AudioSessionStateActive', 'AudioSessionStateExpired');

  cAudioSessionDisconnected: Array [TAudioSessionDisconnected] of String =
    ('device removed', 'server shut down', 'format changed', 'user logged off',
    'session disconnected', 'exclusive-mode override');
type
  IAudioSessionEvents = interface(IUnknown)
    ['{24918ACC-64B3-37C1-8CA9-74A66E9957A8}']
    function OnDisplayNameChanged(NewDisplayName: LPCWSTR; EventContext: pGuid): HResult; stdcall;
    function OnIconPathChanged(NewIconPath: LPCWSTR; EventContext: pGuid): HResult; stdcall;
    function OnSimpleVolumeChanged(NewVolume: Single; NewMute: LongBool; EventContext: pGuid)
      : HResult; stdcall;
    function OnChannelVolumeChanged(ChannelCount: uint; NewChannelArray: PSingle;
      ChangedChannel: uint; EventContext: pGuid): HResult; stdcall;
    function OnGroupingParamChanged(NewGroupingParam, EventContext: pGuid): HResult; stdcall;
    function OnStateChanged(NewState: uint): HResult; stdcall;
    function OnSessionDisconnected(DisconnectReason: uint): HResult; stdcall;
  end;

  TAudioSessionEvents = class(TInterfacedObject, IAudioSessionEvents)
  private
    MsgHandle: HWND;
    VolMute: TVolMute;
  public
    function OnDisplayNameChanged(NewDisplayName: LPCWSTR; EventContext: pGuid): HResult; stdcall;
    function OnIconPathChanged(NewIconPath: LPCWSTR; EventContext: pGuid): HResult; stdcall;
    function OnSimpleVolumeChanged(NewVolume: Single; NewMute: LongBool; EventContext: pGuid)
      : HResult; stdcall;
    function OnChannelVolumeChanged(ChannelCount: uint; NewChannelArray: PSingle;
      ChangedChannel: uint; EventContext: pGuid): HResult; stdcall;
    function OnGroupingParamChanged(NewGroupingParam, EventContext: pGuid): HResult; stdcall;
    function OnStateChanged(NewState: uint): HResult; stdcall;
    function OnSessionDisconnected(DisconnectReason: uint): HResult; stdcall;
    procedure SetMsgHandle(WinHandle: HWND);
  end;

implementation

function TAudioSessionEvents.OnDisplayNameChanged(NewDisplayName: LPCWSTR; EventContext: pGuid)
  : HResult; stdcall;
begin
  Result := S_OK;
end;

function TAudioSessionEvents.OnIconPathChanged(NewIconPath: LPCWSTR; EventContext: pGuid): HResult;
  stdcall;
begin
  Result := S_OK;
end;

function TAudioSessionEvents.OnSimpleVolumeChanged(NewVolume: Single; NewMute: LongBool;
  EventContext: pGuid): HResult; stdcall;
begin
  VolMute.Volume := NewVolume;
  VolMute.Muted := NewMute;
  PostMessage(MsgHandle, WM_VolumeEvent, integer(@VolMute), integer(EventContext));

  Result := S_OK;
end;

function TAudioSessionEvents.OnChannelVolumeChanged(ChannelCount: uint; NewChannelArray: PSingle;
  ChangedChannel: uint; EventContext: pGuid): HResult; stdcall;
begin
  Result := S_OK;
end;

function TAudioSessionEvents.OnGroupingParamChanged(NewGroupingParam, EventContext: pGuid)
  : HResult; stdcall;
begin
  Result := S_OK;
end;

function TAudioSessionEvents.OnStateChanged(NewState: uint): HResult; stdcall;
begin
  PostMessage(MsgHandle, WM_SessionStateEvent, NewState, 0);

  Result := S_OK;
end;

function TAudioSessionEvents.OnSessionDisconnected(DisconnectReason: uint): HResult; stdcall;
begin
  PostMessage(MsgHandle, WM_SessionDisconnected, DisconnectReason, 0);

  Result := S_OK;
end;

procedure TAudioSessionEvents.SetMsgHandle(WinHandle: HWND);
begin
  MsgHandle := WinHandle;
end;

end.
